#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Type
from typing_extensions import Self
from types import TracebackType, FrameType
from multiprocessing.connection import Connection

import time
import timeit
import os
import logging
import multiprocessing
import signal

import psutil

from . import utils, metrics, artifacts
from ...entities import TaskRun
from ...networking import networkManager, NetworkRequestError


def _update(taskRun: TaskRun) -> None:
    logging.getLogger("coretexpylib").debug(">> [Coretex] Heartbeat")
    taskRun.updateStatus()  # updateStatus without params is considered heartbeat

    logging.getLogger("coretexpylib").debug(">> [Coretex] Uploading metrics")
    metrics.upload(taskRun)


def _taskRunWorker(output: Connection, refreshToken: str, taskRunId: int, parentId: int) -> None:
    isStopped = False

    def handleTerminateSignal(signum: int, frame: Optional[FrameType]) -> None:
        if signum != signal.SIGTERM:
            return

        logging.getLogger("coretexpylib").debug(">> [Coretex] Received terminate signal. Terminating...")

        nonlocal isStopped
        isStopped = True

    signal.signal(signal.SIGTERM, handleTerminateSignal)
    utils.initializeLogger(taskRunId)

    response = networkManager.authenticateWithRefreshToken(refreshToken)
    if response.hasFailed():
        utils.sendFailure(output, "Failed to authenticate with refresh token")
        return

    try:
        taskRun: TaskRun = TaskRun.fetchById(taskRunId)
    except NetworkRequestError:
        utils.sendFailure(output, f"Failed to fetch TaskRun with id \"{taskRunId}\"")
        return

    try:
        metrics.create(taskRun)
    except NetworkRequestError:
        utils.sendFailure(output, "Failed to create metrics")
        return

    utils.sendSuccess(output, "TaskRun worker succcessfully started")

    parent = psutil.Process(parentId)
    current = psutil.Process(os.getpid())

    # Start tracking files which are created inside current working directory
    with artifacts.track(taskRun):
        while parent.is_running() and not isStopped:
            logging.getLogger("coretexpylib").debug(f">> [Coretex] Worker process id {current.pid}, parent process id {parent.pid}")

            # Measure elapsed time to calculate for how long should the process sleep
            start = timeit.default_timer()
            _update(taskRun)
            diff = timeit.default_timer() - start

            # Make sure that metrics and heartbeat are sent every 5 seconds
            if diff < 5:
                sleepTime = 5 - diff
                logging.getLogger("coretexpylib").debug(f">> [Coretex] Sleeping for {sleepTime}s")
                time.sleep(sleepTime)

    logging.getLogger("coretexpylib").debug(">> [Coretex] Finished")


class TaskRunWorker:

    def __init__(self, refreshToken: str, taskRunId: int) -> None:
        self._refreshToken = refreshToken

        output, input = multiprocessing.Pipe()
        self.__input = input

        self.__process = multiprocessing.Process(
            name = f"TaskRun {taskRunId} worker process",
            target = _taskRunWorker,
            args = (output, refreshToken, taskRunId, os.getpid()),
            daemon = True
        )

    def start(self) -> None:
        self.__process.start()

        result = self.__input.recv()
        if result["code"] != 0:
            raise RuntimeError(result["message"])

        message = result["message"]
        logging.getLogger("coretexpylib").info(f">> [Coretex] {message}")

    def stop(self) -> None:
        logging.getLogger("coretexpylib").info(">> [Coretex] Stopping the worker process")

        self.__process.terminate()
        self.__process.join()

    def __enter__(self) -> Self:
        self.start()
        return self

    def __exit__(
        self,
        exceptionType: Optional[Type[BaseException]],
        exceptionValue: Optional[BaseException],
        exceptionTraceback: Optional[TracebackType]
    ) -> None:

        if self.__process.is_alive():
            self.stop()

    def kill(self) -> None:
        logging.getLogger("coretexpylib").info(">> [Coretex] Killing the worker process")
        self.__process.kill()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Any
from enum import IntEnum
from pathlib import Path

import requests

from . import config_defaults
from .cron import jobExists, scheduleJob
from .node import getRepoFromImageUrl, getTagFromImageUrl
from ..resources import RESOURCES_DIR
from ...utils import command
from ...configuration import CONFIG_DIR, getInitScript


UPDATE_SCRIPT_NAME = "ctx_node_update.sh"


class NodeStatus(IntEnum):

    inactive     = 1
    active       = 2
    busy         = 3
    deleted      = 4
    reconnecting = 5


def generateUpdateScript(config: Dict[str, Any]) -> str:
    _, dockerPath, _ = command(["which", "docker"], ignoreStdout = True, ignoreStderr = True)
    bashScriptTemplatePath = RESOURCES_DIR / "update_script_template.sh"

    with bashScriptTemplatePath.open("r") as scriptFile:
        bashScriptTemplate = scriptFile.read()

    return bashScriptTemplate.format(
        dockerPath = dockerPath.strip(),
        repository = getRepoFromImageUrl(config["image"]),
        tag = getTagFromImageUrl(config["image"]),
        serverUrl = config["serverUrl"],
        storagePath = config["storagePath"],
        nodeAccessToken = config["nodeAccessToken"],
        nodeMode = config["nodeMode"],
        modelId = config.get("modelId"),
        containerName = config_defaults.DOCKER_CONTAINER_NAME,
        networkName = config_defaults.DOCKER_CONTAINER_NETWORK,
        restartPolicy = "always",
        ports = "21000:21000",
        capAdd = "SYS_PTRACE",
        ram = config["nodeRam"],
        swapMemory = config["nodeSwap"],
        sharedMemory = config["nodeSharedMemory"],
        cpuCount = config.get("cpuCount", config_defaults.DEFAULT_CPU_COUNT),
        imageType = "cpu" if config["allowGpu"] is False else "gpu",
        allowDocker = config.get("allowDocker", False),
        initScript = getInitScript(config),
        nodeSecret = config.get("nodeSecret", config_defaults.DEFAULT_NODE_SECRET)
    )


def dumpScript(updateScriptPath: Path, config: Dict[str, Any]) -> None:
    with updateScriptPath.open("w") as scriptFile:
        scriptFile.write(generateUpdateScript(config))

    command(["chmod", "+x", str(updateScriptPath)], ignoreStdout = True)


def activateAutoUpdate(configDir: Path, config: Dict[str, Any]) -> None:
    updateScriptPath = CONFIG_DIR / UPDATE_SCRIPT_NAME
    dumpScript(updateScriptPath, config)

    if not jobExists(UPDATE_SCRIPT_NAME):
        scheduleJob(configDir, UPDATE_SCRIPT_NAME)


def getNodeStatus() -> NodeStatus:
    try:
        response = requests.get(f"http://localhost:21000/status", timeout = 1)
        status = response.json()["status"]
        return NodeStatus(status)
    except:
        return NodeStatus.inactive



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Final, List, Set
from enum import Enum
from abc import ABC, abstractmethod

import logging

from ..annotation import CoretexImageAnnotation
from ...entities import ImageDataset, ImageDatasetClass
from ...threading import MultithreadedDataProcessor


class ConverterProcessorType(Enum):

    """
        List of all annotation formats supported by Coretex
    """

    coco = 0
    yolo = 1
    createML = 2
    voc = 3
    labelMe = 4
    pascalSeg = 5

    # TODO: Migrate to Human Segmentation task template repo, or try to make it generic
    humanSegmentation = 6

    cityScape = 7


class BaseConverter(ABC):

    """
        Base class for Coretex Annotation format conversion

        Properties
        ----------
        datasetName : str
            name of dataset
        projectId : int
            id of Coretex Project
        datasetPath : str
            path to dataset
    """

    def __init__(self, datasetName: str, projectId: int, datasetPath: str) -> None:
        self._dataset: Final = ImageDataset.createDataset(datasetName, projectId)
        self._datasetPath: Final = datasetPath

    def _saveImageAnnotationPair(self, imagePath: str, annotation: CoretexImageAnnotation) -> None:
        sample = self._dataset.add(imagePath)

        # Attach annotation to sample
        if not sample.saveAnnotation(annotation):
            logging.getLogger("coretexpylib").info(">> [Coretex] Failed to save ImageSample annotation")

    @abstractmethod
    def _dataSource(self) -> List[Any]:
        pass

    @abstractmethod
    def _extractLabels(self) -> Set[str]:
        pass

    @abstractmethod
    def _extractSingleAnnotation(self, value: Any) -> None:
        pass

    def convert(self) -> ImageDataset:
        """
            Converts the dataset to Coretex Format

            Returns
            -------
            ImageDatasetType -> The converted ImageDataset object
        """

        # Extract classes
        labels = self._extractLabels()
        classes = ImageDatasetClass.generate(labels)

        if self._dataset.saveClasses(classes):
            logging.getLogger("coretexpylib").info(">> [Coretex] Dataset classes saved successfully")
        else:
            logging.getLogger("coretexpylib").info(">> [Coretex] Failed to save dataset classes")

        # Extract annotations
        MultithreadedDataProcessor(
            self._dataSource(),
            self._extractSingleAnnotation,
            message = "Converting dataset..."
        ).process()

        if not self._dataset.finalize():
            raise ValueError(f"Failed to finalize dataset \"{self._dataset.name}\"")

        return self._dataset



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Any, Dict, List, Union, Tuple
from pathlib import Path
from abc import ABC, abstractmethod
from contextlib import ExitStack
from http import HTTPStatus
from importlib.metadata import version as getLibraryVersion

import os
import json
import logging
import platform
import random
import time

import requests
import requests.adapters

from .utils import RequestBodyType, RequestFormType, logFilesData, logRequestFailure
from .request_type import RequestType
from .network_response import NetworkResponse, NetworkRequestError
from .file_data import FileData


REQUEST_TIMEOUT = 5
MAX_RETRY_COUNT = 5
LOGIN_ENDPOINT = "user/login"
REFRESH_ENDPOINT = "user/refresh"
API_TOKEN_HEADER = "api-token"
API_TOKEN_KEY = "token"
REFRESH_TOKEN_KEY = "refresh_token"

RETRY_STATUS_CODES = [
    HTTPStatus.TOO_MANY_REQUESTS,
    HTTPStatus.INTERNAL_SERVER_ERROR,
    HTTPStatus.SERVICE_UNAVAILABLE
]

TimeoutType = Optional[Union[int, Tuple[int, int]]]


def getDelayBeforeRetry(retryCount: int) -> int:
    # retryCount starts from 0 so we add +1/+2 to start/end
    # to make it have proper delay

    start = (retryCount + 1) ** 2
    end   = (retryCount + 2) ** 2

    return random.randint(start, end)


class RequestFailedError(Exception):

    def __init__(self, endpoint: str, type_: RequestType) -> None:
        super().__init__(f">> [Coretex] \"{type_.name}\" request failed for endpoint \"{endpoint}\"")


class NetworkManagerBase(ABC):

    def __init__(self) -> None:
        self._session = requests.Session()

        cpuCount = os.cpu_count()
        if cpuCount is None:
            cpuCount = 1

        # 10 is default, keep that value for machines which have <= 10 cores
        adapter = requests.adapters.HTTPAdapter(pool_maxsize = max(10, cpuCount))
        self._session.mount("http://", adapter)
        self._session.mount("https://", adapter)

    @property
    def serverUrl(self) -> str:
        return os.environ["CTX_API_URL"] + "api/v1/"

    @property
    @abstractmethod
    def _apiToken(self) -> Optional[str]:
        pass

    @_apiToken.setter
    @abstractmethod
    def _apiToken(self, value: Optional[str]) -> None:
        pass

    @property
    @abstractmethod
    def _refreshToken(self) -> Optional[str]:
        pass

    @_refreshToken.setter
    @abstractmethod
    def _refreshToken(self, value: Optional[str]) -> None:
        pass

    @property
    def userAgent(self) -> str:
        coretexpylibVersion = getLibraryVersion("coretex")
        return f"coretexpylib;{coretexpylibVersion};python;{platform.python_version()}"

    @property
    def hasStoredCredentials(self) -> bool:
        """
            To use this functions call it using coretex.networking.networkManager

            Raises
            ------
            NotImplementedError
        """

        raise NotImplementedError

    def _headers(self, contentType: str = "application/json") -> Dict[str, str]:
        headers = {
            "Content-Type": contentType,
            "Connection": "keep-alive",
            "X-User-Agent": self.userAgent
        }

        if self._apiToken is not None:
            headers[API_TOKEN_HEADER] = self._apiToken

        return headers

    def shouldRetry(self, retryCount: int, response: Optional[NetworkResponse]) -> bool:
        """
            Checks if network request should be repeated based on the number of repetitions
            as well as the response from previous repetition

            Parameters
            ----------
            retryCount : int
                number of repeated function calls
            response : Optional[NetworkResponse]
                response of the request which is pending for retry

            Returns
            -------
            bool -> True if the request should be retried, False if not
        """

        # Limit retry count to 3 times
        if retryCount >= MAX_RETRY_COUNT:
            return False

        if response is not None:
            # If we get unauthorized maybe API token is expired
            # If refresh endpoint failed with unauthorized do not retry
            if response.isUnauthorized() and response.endpoint != REFRESH_ENDPOINT:
                refreshTokenResponse = self.refreshToken()
                return not refreshTokenResponse.hasFailed()

            return response.statusCode in RETRY_STATUS_CODES

        return True

    def request(
        self,
        endpoint: str,
        requestType: RequestType,
        headers: Optional[Dict[str, str]] = None,
        query: Optional[Dict[str, Any]] = None,
        body: Optional[RequestBodyType] = None,
        files: Optional[RequestFormType] = None,
        auth: Optional[Tuple[str, str]] = None,
        timeout: Optional[TimeoutType] = REQUEST_TIMEOUT,
        stream: bool = False,
        retryCount: int = 0
    ) -> NetworkResponse:

        """
            Sends an HTTP request with provided parameters
            This method is used as a base for all other networkManager methods

            Parameters
            ----------
            endpoint : str
                endpoint to which the request is sent
            requestType : RequestType
                type of the request which is sent (get, post, put, delete, etc...)
            headers : Optional[Dict[str, Any]]
                headers which will be sent with request, if None default values will be used
            query : Optional[Dict[str, Any]]
                parameters which will be sent as query parameters
            body : Optional[RequestBodyType]
                parameters which will be sent as request body
            files : Optional[RequestFormType]
                files which will be sent as a part of form data request
            auth : Optional[Tuple[str, str]]
                credentials which will be send as basic auth header
            stream : bool
                defines if request body will be downloaded as a stream or not
            retryCount : int
                retry number of request - only for internal use

            Returns
            -------
            NetworkResponse -> object containing the request response

            Raises
            ------
            RequestFailedError -> if request failed due to connection issues
        """

        if headers is None:
            headers = self._headers()

        url = self.serverUrl + endpoint

        # Log request debug data
        logging.getLogger("coretexpylib").debug(f">> [Coretex] Sending request to \"{url}\"")
        logging.getLogger("coretexpylib").debug(f"\tType: {requestType}")
        logging.getLogger("coretexpylib").debug(f"\tHeaders: {headers}")
        logging.getLogger("coretexpylib").debug(f"\tQuery: {query}")
        logging.getLogger("coretexpylib").debug(f"\tBody: {body}")
        logging.getLogger("coretexpylib").debug(f"\tFiles: {logFilesData(files)}")
        logging.getLogger("coretexpylib").debug(f"\tAuth: {auth}")
        logging.getLogger("coretexpylib").debug(f"\tStream: {stream}")
        logging.getLogger("coretexpylib").debug(f"\tRetry count: {retryCount}")

        # If Content-Type is application/json make sure that body is converted to json
        data: Optional[Any] = body
        if headers.get("Content-Type") == "application/json" and data is not None:
            data = json.dumps(body)

        try:
            rawResponse = self._session.request(
                requestType.value,
                url,
                params = query,
                data = data,
                auth = auth,
                timeout = timeout,
                files = files,
                headers = headers
            )

            response = NetworkResponse(rawResponse, endpoint)
            if response.hasFailed():
                logRequestFailure(endpoint, response)

            if self.shouldRetry(retryCount, response):
                if self._apiToken is not None:
                    headers[API_TOKEN_HEADER] = self._apiToken

                # If we hit rate limiter sleep before retrying the request
                if response.statusCode == HTTPStatus.TOO_MANY_REQUESTS:
                    delay = getDelayBeforeRetry(retryCount)
                    logging.getLogger("coretexpylib").debug(f">> [Coretex] Waiting for {delay} seconds before retrying failed \"{endpoint}\" request")

                    time.sleep(delay)

                return self.request(endpoint, requestType, headers, query, body, files, auth, timeout, stream, retryCount + 1)

            return response
        except BaseException as exception:
            logging.getLogger("coretexpylib").debug(f">> [Coretex] Request failed. Reason \"{exception}\"", exc_info = exception)

            if self.shouldRetry(retryCount, None):
                if self._apiToken is not None:
                    headers[API_TOKEN_HEADER] = self._apiToken

                return self.request(endpoint, requestType, headers, query, body, files, auth, timeout, stream, retryCount + 1)

            raise RequestFailedError(endpoint, requestType)

    def post(self, endpoint: str, params: Optional[RequestBodyType] = None) -> NetworkResponse:
        """
            Sends post HTTP request

            Parameters
            ----------
            endpoint : str
                endpoint to which the request is sent
            params : Optional[RequestBodyType]
                body of the request

            Returns
            -------
            NetworkResponse -> object containing the request response

            Raises
            ------
            RequestFailedError -> if request failed due to connection issues
        """

        return self.request(endpoint, RequestType.post, body = params)

    def get(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> NetworkResponse:
        """
            Sends get HTTP request

            Parameters
            ----------
            endpoint : str
                endpoint to which the request is sent
            params : Optional[RequestBodyType]
                query parameters of the request

            Returns
            -------
            NetworkResponse -> object containing the request response

            Raises
            ------
            RequestFailedError -> if request failed due to connection issues
        """

        return self.request(endpoint, RequestType.get, query = params)

    def put(self, endpoint: str, params: Optional[RequestBodyType] = None) -> NetworkResponse:
        """
            Sends put HTTP request

            Parameters
            ----------
            endpoint : str
                endpoint to which the request is sent
            params : Optional[RequestBodyType]
                body of the request

            Returns
            -------
            NetworkResponse -> object containing the request response

            Raises
            ------
            RequestFailedError -> if request failed due to connection issues
        """

        return self.request(endpoint, RequestType.put, body = params)

    def delete(self, endpoint: str) -> NetworkResponse:
        """
            Sends delete HTTP request

            Parameters
            ----------
            endpoint : str
                endpoint to which the request is sent

            Returns
            -------
            NetworkResponse -> object containing the request response

            Raises
            ------
            RequestFailedError -> if request failed due to connection issues
        """

        return self.request(endpoint, RequestType.delete)

    def formData(
        self,
        endpoint: str,
        params: Optional[Dict[str, Any]] = None,
        files: Optional[List[FileData]] = None,
        timeout: TimeoutType = REQUEST_TIMEOUT
    ) -> NetworkResponse:

        """
            Sends multipart/form-data request

            Parameters
            ----------
            endpoint : str
                endpoint to which the request is sent
            params : Optional[Dict[str, Any]]
                form data parameters
            files : Optional[List[FileData]]
                form data files
            timeout : Optional[Union[int, Tuple[int, int]]]
                timeout for the request

            Returns
            -------
            NetworkResponse -> object containing the request response

            Example
            -------
            >>> from coretex import networkManager
            \b
            >>> response = networkManager.formData(
                    endpoint = "dummyObject/form",
                    params = {
                        "key": "value"
                    }
                )
            >>> if response.hasFailed():
                    print("Failed to send form data request")
        """

        if files is None:
            files = []

        with ExitStack() as stack:
            filesData = [file.prepareForUpload(stack) for file in files]

            headers = self._headers("multipart/form-data")
            del headers["Content-Type"]

            if len(files) > 0:
                response = self.request(endpoint, RequestType.options, timeout = REQUEST_TIMEOUT)
                if response.hasFailed():
                    raise NetworkRequestError(response, "Could not establish a connection with the server")

                timeout = None

            return self.request(endpoint, RequestType.post, headers, body = params, files = filesData, timeout = timeout)

        # mypy is complaining about missing return statement but this code is unreachable
        # see: https://github.com/python/mypy/issues/7726
        raise RuntimeError("Unreachable")

    def authenticate(self, username: str, password: str, storeCredentials: bool = True) -> NetworkResponse:
        """
            Authenticates user with provided credentials

            Parameters
            ----------
            username : str
                Coretex.ai username
            password : str
                Coretex.ai password
            storeCredentials : bool
                If true credentials will be stored in User object for reuse,
                ignored for all managers except coretex.networking.networkManager

            Returns
            -------
            NetworkResponse -> object containing the request response

            Example
            -------
            >>> from coretex.networking import networkManager
            \b
            >>> response = networkManager.authenticate("dummy@coretex.ai", "123456")
            >>> if response.hasFailed():
                    print("Failed to authenticate")
        """

        # authenticate using credentials stored in requests.Session.auth

        response = self.request(LOGIN_ENDPOINT, RequestType.post, auth = (username, password))
        if response.hasFailed():
            return response

        responseJson = response.getJson(dict)

        self._apiToken = responseJson[API_TOKEN_KEY]
        self._refreshToken = responseJson[REFRESH_TOKEN_KEY]

        return response

    def authenticateWithStoredCredentials(self) -> NetworkResponse:
        """
            To use this functions call it using coretex.networking.networkManager

            Raises
            ------
            NotImplementedError
        """

        raise NotImplementedError

    def authenticateWithRefreshToken(self, token: str) -> NetworkResponse:
        """
            Authenticates user with provided refresh token

            Parameters
            ----------
            token : str
                refresh token

            Returns
            -------
            NetworkResponse -> object containing the request response
        """

        self._refreshToken = token
        return self.refreshToken()

    def download(
        self,
        endpoint: str,
        destination: Union[Path, str],
        params: Optional[Dict[str, Any]] = None
    ) -> NetworkResponse:

        """
            Downloads file to the given destination

            Parameters
            ----------
            endpoint : str
                endpoint to which the request is sent
            destination : Union[Path, str]
                path to save file

            Returns
            -------
            NetworkResponse -> object containing the request response

            Example
            -------
            >>> from coretex import networkManager
            \b
            >>> response = networkManager.genericDownload(
                    endpoint = "dummyObject/download",
                    destination = "path/to/destination/folder"
                )
            >>> if response.hasFailed():
                    print("Failed to download the file")
        """

        if isinstance(destination, str):
            destination = Path(destination)

        response = self.get(endpoint, params)
        if response.hasFailed():
            return response

        with destination.open("wb") as downloadedFile:
            downloadedFile.write(response.getContent())

        return response

    def streamDownload(
        self,
        endpoint: str,
        destination: Union[Path, str],
        params: Optional[Dict[str, Any]] = None
    ) -> NetworkResponse:

        """
            Downloads file to the given destination

            Parameters
            ----------
            endpoint : str
                endpoint to which the request is sent
            destination : Union[Path, str]
                path to save file
            retryCount : int
                number of function calls if request has failed, used
                for internal retry mechanism

            Returns
            -------
            NetworkResponse -> object containing the request response

            Example
            -------
            >>> from coretex import networkManager
            \b
            >>> response = networkManager.streamDownload(
                    endpoint = "dummyObject/download",
                    destination = "path/to/destination/folder"
                )
            >>> if response.hasFailed():
                    print("Failed to download the file")
        """

        if isinstance(destination, str):
            destination = Path(destination)

        response = self.request(endpoint, RequestType.get, query = params, stream = True, timeout = (5, 60))
        if response.hasFailed():
            return response

        if destination.exists() and "Content-Length" in response.headers:
            contentLength = int(response.headers["Content-Length"])
            if contentLength == destination.stat().st_size:
                return response

            destination.unlink()

        with destination.open("wb") as file:
            for chunk in response.stream():
                file.write(chunk)

        return response

    def refreshToken(self) -> NetworkResponse:
        """
            Uses refresh token functionality to fetch new API access token

            Returns
            -------
            NetworkResponse -> object containing the request response
        """

        if self._refreshToken is None:
            raise ValueError(f">> [Coretex] Cannot send \"{REFRESH_ENDPOINT}\" request, refreshToken is None")

        headers = self._headers()
        headers[API_TOKEN_HEADER] = self._refreshToken

        response = self.request(REFRESH_ENDPOINT, RequestType.post, headers = headers)
        if response.hasFailed():
            return response

        responseJson = response.getJson(dict)
        self._apiToken = responseJson[API_TOKEN_KEY]

        return response

    def reset(self) -> None:
        """
            Removes api and refresh token
        """

        self._apiToken = None
        self._refreshToken = None



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Dict, Tuple, Optional
from typing_extensions import Self
from uuid import UUID
from math import cos, sin, radians

from PIL import Image, ImageDraw

import numpy as np

from .bbox import BBox
from .classes_format import ImageDatasetClasses
from ....codable import Codable, KeyDescriptor


SegmentationType = List[int]


def toPoly(segmentation: List[int]) -> List[Tuple[int, int]]:
    points: List[Tuple[int, int]] = []

    for index in range(0, len(segmentation) - 1, 2):
        points.append((segmentation[index], segmentation[index + 1]))

    return points


class CoretexSegmentationInstance(Codable):

    """
        Segmentation Instance class

        Properties
        ----------
        classID : UUID
            uuid of class
        bbox : BBox
            Bounding Box as a python class
        segmentations : List[SegmentationType]
            list of segmentations that define the precise boundaries of object
    """

    classId: UUID
    bbox: BBox
    segmentations: List[SegmentationType]

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()

        descriptors["classId"] = KeyDescriptor("class_id", UUID)
        descriptors["bbox"] = KeyDescriptor("bbox", BBox)
        descriptors["segmentations"] = KeyDescriptor("annotations")

        return descriptors

    @classmethod
    def create(cls, classId: UUID, bbox: BBox, segmentations: List[SegmentationType]) -> Self:
        """
            Creates CoretexSegmentationInstance object with provided parameters

            Parameters
            ----------
            classID : UUID
                uuid of class
            bbox : BBox
                Bounding Box as a python class
            segmentations : List[SegmentationType]
                list of segmentations that define the precise boundaries of object

            Returns
            -------
            The created CoretexSegmentationInstance object
        """

        obj = cls()

        obj.classId = classId
        obj.bbox = bbox
        obj.segmentations = segmentations

        return obj

    def extractSegmentationMask(self, width: int, height: int) -> np.ndarray:
        """
            Generates segmentation mask based on provided
            width and height of image\n
            Pixel values are equal to class IDs

            Parameters
            ----------
            width : int
                width of image in pixels
            height : int
                height of image in pixels

            Returns
            -------
            np.ndarray -> segmentation mask represented as np.ndarray
        """

        image = Image.new("L", (width, height))

        for segmentation in self.segmentations:
            draw = ImageDraw.Draw(image)
            draw.polygon(toPoly(segmentation), fill = 1)

        return np.array(image)

    def extractBinaryMask(self, width: int, height: int) -> np.ndarray:
        """
            Works the same way as extractSegmentationMask function
            Values that are > 0 are capped to 1

            Parameters
            ----------
            width : int
                width of image in pixels
            height : int
                height of image in pixels

            Returns
            -------
            np.ndarray -> binary segmentation mask represented as np.ndarray
        """

        binaryMask = self.extractSegmentationMask(width, height)
        binaryMask[binaryMask > 0] = 1

        return binaryMask

    def centroid(self) -> Tuple[int, int]:
        """
            Calculates centroid of segmentations

            Returns
            -------
            Tuple[int, int] -> x, y coordinates of centroid
        """

        flattenedSegmentations = [element for sublist in self.segmentations for element in sublist]

        listCX = [value for index, value in enumerate(flattenedSegmentations) if index % 2 == 0]
        centerX = sum(listCX) // len(listCX)

        listCY = [value for index, value in enumerate(flattenedSegmentations) if index % 2 != 0]
        centerY = sum(listCY) // len(listCY)

        return centerX, centerY

    def centerSegmentations(self, newCentroid: Tuple[int, int]) -> None:
        """
            Centers segmentations to the specified center point

            Parameters
            ----------
            newCentroid : Tuple[int, int]
                x, y coordinates of centroid
            """

        newCenterX, newCenterY = newCentroid
        oldCenterX, oldCenterY = self.centroid()

        modifiedSegmentations: List[List[int]] = []

        for segmentation in self.segmentations:
            modifiedSegmentation: List[int] = []

            for i in range(0, len(segmentation), 2):
                x = segmentation[i] + (newCenterX - oldCenterX)
                y = segmentation[i+1] + (newCenterY - oldCenterY)

                modifiedSegmentation.append(x)
                modifiedSegmentation.append(y)

            modifiedSegmentations.append(modifiedSegmentation)

        self.segmentations = modifiedSegmentations

    def rotateSegmentations(
        self,
        degrees: int,
        origin: Optional[Tuple[int, int]] = None
    ) -> None:

        """
            Rotates segmentations of CoretexSegmentationInstance object

            Parameters
            ----------
            degrees : int
                degree of rotation
        """

        if origin is None:
            origin = self.centroid()

        rotatedSegmentations: List[List[int]] = []
        centerX, centerY = origin

        # because rotations with image and segmentations doesn't go in same direction
        # one of the rotations has to be inverted so they go in same direction
        theta = radians(-degrees)
        cosang, sinang = cos(theta), sin(theta)

        for segmentation in self.segmentations:
            rotatedSegmentation: List[int] = []

            for i in range(0, len(segmentation), 2):
                x = segmentation[i] - centerX
                y = segmentation[i + 1] - centerY

                newX = int(x * cosang - y * sinang) + centerX
                newY = int(x * sinang + y * cosang) + centerY

                rotatedSegmentation.append(newX)
                rotatedSegmentation.append(newY)

            rotatedSegmentations.append(rotatedSegmentation)

        self.segmentations = rotatedSegmentations


class CoretexImageAnnotation(Codable):

    """
        Image Annotation class

        Properties
        ----------
        name : str
            name of annotation class
        width : int
            width of annotation
        height : int
            height of annotation
        instances : List[CoretexSegmentationInstance]
            list of SegmentationInstance objects
    """

    name: str
    width: int
    height: int
    instances: List[CoretexSegmentationInstance]

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["instances"] = KeyDescriptor("instances", CoretexSegmentationInstance, list)

        return descriptors

    @classmethod
    def create(
        cls,
        name: str,
        width: int,
        height: int,
        instances: List[CoretexSegmentationInstance]
    ) -> Self:
        """
            Creates CoretexImageAnnotation object with provided parameters

            Parameters
            ----------
            name : str
                name of annotation class
            width : int
                width of annotation
            height : int
                height of annotation
            instances : List[CoretexSegmentationInstance]
                list of SegmentationInstance objects

            Returns
            -------
            The created CoretexImageAnnotation object
        """

        obj = cls()

        obj.name = name
        obj.width = width
        obj.height = height
        obj.instances = instances

        return obj

    def extractSegmentationMask(self, classes: ImageDatasetClasses) -> np.ndarray:
        """
            Generates segmentation mask of provided ImageDatasetClasses object

            Parameters
            ----------
            classes : ImageDatasetClasses
                list of dataset classes

            Returns
            -------
            np.ndarray -> segmentation mask represented as np.ndarray
        """

        image = Image.new("L", (self.width, self.height))

        for instance in self.instances:
            labelId = classes.labelIdForClassId(instance.classId)
            if labelId is None:
                continue

            for segmentation in instance.segmentations:
                if len(segmentation) == 0:
                    raise ValueError(f">> [Coretex] Empty segmentation")

                draw = ImageDraw.Draw(image)
                draw.polygon(toPoly(segmentation), fill = labelId + 1)

        return np.asarray(image)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, List, Set

import os
import re
import logging

from PIL import Image

from ..base_converter import BaseConverter
from ...annotation import CoretexImageAnnotation, CoretexSegmentationInstance, BBox


class Helper:

    @staticmethod
    def isFloat(value: str) -> bool:
        try:
            float(value)
            return True
        except ValueError:
            return False


class YoloConverter(BaseConverter):

    def __init__(self, datasetName: str, projectId: int, datasetPath: str) -> None:
        super().__init__(datasetName, projectId, datasetPath)

        self.__imagesPath = os.path.join(datasetPath, "images")
        self.__annotations = os.path.join(datasetPath, "annotations")

        classesPath = os.path.join(self.__annotations, "classes.txt")
        if not os.path.exists(classesPath):
            raise FileNotFoundError(">> [Coretex] classes.txt file not found")

        with open(classesPath, 'r') as f:
            text = f.read()
            self.__rawLabels = text.split("\n")
            self.__rawLabels = [label for label in self.__rawLabels if label]

    def _dataSource(self) -> List[str]:
        return os.listdir(self.__annotations)

    def _extractLabels(self) -> Set[str]:
        return set(self.__rawLabels)

    def __extractBBox(self, rawInstance: List[str], width: int, height: int) -> BBox:
        xYolo = float(rawInstance[1])
        yYolo = float(rawInstance[2])
        wYolo = float(rawInstance[3])
        hYolo = float(rawInstance[4])

        boxWidth = int(wYolo * width)
        boxHeight = int(hYolo * height)
        xMin = int(xYolo * width - (boxWidth / 2))
        yMin = int(yYolo * height - (boxHeight / 2))

        return BBox(xMin, yMin, boxWidth, boxHeight)

    def __extractInstance(self, rawInstance: List[str], width: int, height: int) -> Optional[CoretexSegmentationInstance]:
        # Get class name
        labelId = int(rawInstance[0])
        label = self.__rawLabels[labelId]

        coretexClass = self._dataset.classByName(label)
        if coretexClass is None:
            logging.getLogger("coretexpylib").info(f">> [Coretex] Class: ({label}) is not a part of dataset")
            return None

        bbox = self.__extractBBox(rawInstance, width, height)
        return CoretexSegmentationInstance.create( coretexClass.classIds[0], bbox, [bbox.polygon])

    def _extractSingleAnnotation(self, yoloFilePath: str) -> None:
        if not yoloFilePath.endswith("txt"):
            return

        if os.path.splitext(yoloFilePath)[0] == "classes":
            return

        yoloFilePath = os.path.join(self.__annotations, yoloFilePath)
        yoloName = os.path.basename(yoloFilePath)

        imagePath = self.imageCheck(os.path.join(self.__imagesPath, yoloName))
        if imagePath is not None:
            imageName = os.path.basename(imagePath)

        baseImageName = os.path.splitext(imageName)[0]
        baseYoloName = os.path.splitext(yoloName)[0]

        if baseImageName != baseYoloName:
            return

        with open(yoloFilePath, 'r') as file:
            allLines = file.readlines()

            image = Image.open(imagePath)
            coretexAnnotation = CoretexImageAnnotation.create(imageName, image.width, image.height, [])

            # Get bounding boxes and classes from yolo txt
            for line in allLines:
                yoloArray = re.split("\s", line.rstrip())
                isFormatCorrect = YoloConverter.formatCheck(yoloArray)

                if not isFormatCorrect:
                    continue

                instance = self.__extractInstance(yoloArray, image.width, image.height)
                if instance is None:
                    continue

                coretexAnnotation.instances.append(instance)

            self._saveImageAnnotationPair(os.path.join(self.__imagesPath, imageName), coretexAnnotation)

    @staticmethod
    def imageCheck(yoloFilePath: str) -> Optional[str]:
        if os.path.exists(yoloFilePath.replace('txt', 'jpeg')):
            return yoloFilePath.replace('txt', 'jpeg')
        if os.path.exists(yoloFilePath.replace('txt', 'jpg')):
            return yoloFilePath.replace('txt', 'jpg')
        if os.path.exists(yoloFilePath.replace('txt', 'png')):
           return yoloFilePath.replace('txt', 'png')

        return None

    @staticmethod
    def formatCheck(yoloArray: List[str]) -> bool:
        """
            Checks format of yolo annotation file

            Parameters
            ----------
            yoloArray : List[str]
                list with label id and bounding boxes

            Returns
            -------
            bool -> True if format is correct, False if format is not correct
        """

        if len(yoloArray) != 5:
            return False

        for value in yoloArray:
            if not Helper.isFloat(value):
                return False

        return True



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, List, Dict, Optional

import json
import logging

from ..base_parameter import BaseParameter


class IMUVectorsParameter(BaseParameter[Dict[str, int]]):

    @property
    def types(self) -> List[type]:
        return [dict]

    def overrideValue(self, value: Optional[Any]) -> Optional[Any]:
        if value is None:
            return None

        try:
            parsedValue = json.loads(value.replace("'", "\""))
            return parsedValue
        except ValueError as e:
            logging.getLogger("coretexpylib").warning(f">> [Coretex] Failed to override IMU vectors parameter \"{self.name}\". | {e}")
            return self.value



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Union
from pathlib import Path

import os
import shutil
import pickle
import hashlib

import requests

from . import folder_manager


class CacheException(Exception):

    """
        Exception which is raised due to any unexpected behaviour with Cache module
    """

    pass


def _hashCacheKey(key: str) -> str:
    return hashlib.sha256(key.encode("UTF-8")).hexdigest()


MAX_RETRY_COUNT = 3


def _download(source: str, destination: Path, retryCount: int = 0) -> Path:
    if destination.exists():
        raise FileExistsError(destination)

    try:
        with requests.get(source, stream = True) as r:
            r.raise_for_status()

            with destination.open("wb") as destinationFile:
                for chunk in r.iter_content(chunk_size = 8192):
                    destinationFile.write(chunk)
    except:
        if retryCount >= MAX_RETRY_COUNT:
            raise

        destination.unlink(missing_ok = True)
        return _download(source, destination, retryCount + 1)

    return destination


def getPath(key: str) -> Path:
    """
        Retrieves the path of the cache

        Parameters
        ----------
        key : str
            key to which the cache item was linked

        Returns
        -------
        pathlib.Path -> path of the cached item

        Raises
        ------
        CacheException -> if something went wrong

        Example
        -------
        >>> from coretex import cache
        \b
        >>> key = "dummyFile"
        >>> print(cache.getPath(key))
        Path("/Users/dummyUser/.coretex/cache/c147efcfc2d7ea666a9e4f5187b115c90903f0fc896a56df9a6ef5d8f3fc9f31.zip")
    """

    if not exists(key):
        raise CacheException(">> [Coretex] Cache with given key doesn't exist.")

    return folder_manager.cache / _hashCacheKey(key)


def exists(key: str) -> bool:
    """
        Checks if the cache item exists

        Parameters
        ----------
        key : str
            key to which the cache item was linked

        Returns
        -------
        bool -> True if it exists, False otherwise

        Example
        -------
        >>> from coretex import cache
        \b
        >>> key = "dummyFile"
        >>> print(cache.exists(key))
        True
    """

    return folder_manager.cache.joinpath(_hashCacheKey(key)).exists()


def remove(key: str) -> None:
    """
        Removes cached item from the cache

        Parameters
        ----------
        key : str -> Key to which the cache item was linked

        Raises
        ------
        CacheException -> if the cache item does not exist

        Example
        -------
        >>> from coretex import cache
        \b
        >>> cache.remove("dummyFile")
    """

    if not exists(key):
        raise CacheException(">> [Coretex] Cache with given key doesn't exist.")

    getPath(key).unlink()


def clear() -> None:
    """
        Clears all cached items

        Example
        -------
        >>> from coretex import cache
        \b
        >>> cache.clear()
    """

    shutil.rmtree(folder_manager.cache)


def storeObject(key: str, object: Any, override: bool = False) -> None:
    """
        Caches the specified object using pickle

        Parameters
        ----------
        key : str
            key to which the cached object will be linked
        object : Any
            object which supports being pickled
        override : bool
            should the cache be overriden if it exists or not

        Raises
        ------
        CacheException -> if something went wrong

        Example
        -------
        >>> from coretex import cache
        \b
        >>> dummyObject = {"name": "John", "age": 24, "gender": "Male"}
        >>> cache.storeObject("Person", dummyObject)
    """

    if not override and exists(key):
        raise CacheException(">> [Coretex] Cache with given key already exists. Set \"override\" to \"True\" if you want to override existing cache.")

    if override and exists(key):
        getPath(key).unlink()

    cachePath = folder_manager.cache / _hashCacheKey(key)
    with cachePath.open("wb") as cacheFile:
        pickle.dump(object, cacheFile)


def storeFile(key: str, source: Union[Path, str], override: bool = False) -> None:
    """
        Caches the specified file

        Parameters
        ----------
        key : str
            key to which the cached file will be linked
        source : str
            path to the file which will be cached
        override : bool
            should the cache be overriden if it exists or not

        Raises
        ------
        CacheException -> if something went wrong

        Example
        -------
        >>> from coretex import cache
        \b
        >>> filePath = "path/to/file"
        >>> cache.storeFile("dummyFile", filePath)
    """

    if not isinstance(source, Path):
        source = Path(source)

    if not override and exists(key):
        raise CacheException(">> [Coretex] Cache with given key already exists. Set \"override\" to \"True\" if you want to override existing cache.")

    if override and exists(key):
        getPath(key).unlink()

    cachePath = folder_manager.cache / _hashCacheKey(key)

    # Hardlink the file to the cache directory
    os.link(source, cachePath)


def storeUrl(key: str, url: str, override: bool = False) -> None:
    """
        Downloads and caches file from the specified URL

        Parameters
        ----------
        url : str
            URL of the file which is downloaded
        fileName : str
            Name of the downloaded file with extension - used as a key for cache

        Returns
        -------
        Tuple[Path, str] -> Path of the cached file and the name of the cached file

        Raises
        ------
        CacheException -> if something went wrong

        Example
        -------
        >>> from coretex import cache
        \b
        >>> url = "https://dummy_url.com/download"
        >>> fileName = "dummyFile.ext"
        >>> cache.storeUrl(url, fileName)
    """

    if not override and exists(key):
        raise CacheException(">> [Coretex] Cache with given key already exists. Set \"override\" to \"True\" if you want to override existing cache.")

    if override and exists(key):
        getPath(key).unlink()

    cachePath = folder_manager.cache / _hashCacheKey(key)
    _download(url, cachePath)


def loadObject(key: str) -> Any:
    """
        Loads the object cached with storeObject function

        Parameters
        ----------
        key : str
            key to which the object was linked when cached

        Returns
        -------
        Any -> unpickled object

        Raises
        ------
        CacheException -> if something went wrong

        Example
        -------
        >>> from coretex import cache
        \b
        >>> loadedObject = cache.load("dummyObject")
        >>> print(loadedObject)
        {"name": "John", "age": 24, "gender": "Male"}
    """

    if not exists(key):
        raise CacheException(">> [Coretex] Cache with given key doesn't exist.")

    with getPath(key).open("rb") as pickleFile:
        return pickle.load(pickleFile)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, List, Optional, Tuple, Dict

import logging

from ..base_parameter import BaseParameter
from ..utils import validateEnumStructure
from ....project import ProjectType


class EnumParameter(BaseParameter[Dict[str, Any]]):

    @property
    def types(self) -> List[type]:
        return NotImplemented

    def validate(self) -> Tuple[bool, Optional[str]]:
        isValid, message = validateEnumStructure(self.name, self.value, self.required)
        if not isValid:
            return isValid, message

        # validateEnumStructure already checks if value is of correct type
        value: Dict[str, Any] = self.value  # type: ignore[assignment]

        selected = value["selected"]
        options = value["options"]

        if selected is None and not self.required:
            return True, None

        if not type(selected) is int:
            return False, f"Enum parameter \"{self.name}.selected\" has invalid type. Expected \"int\", got \"{type(selected).__name__}\""

        if selected >= len(options) or selected < 0:
            return False, f"Enum parameter \"{self.name}.selected\" has out of range value"

        return True, None

    def parseValue(self, type_: ProjectType) -> Optional[Any]:
        if self.value is None:
            return self.value

        selected: Optional[int] = self.value.get("selected")
        if selected is None:
            return None

        return self.value["options"][selected]

    def overrideValue(self, value: Optional[Any]) -> Optional[Any]:
        if value is None or self.value is None:
            return None

        try:
            parsedValue: Dict[str, Any] = self.value
            parsedValue["selected"] = int(value)
            return parsedValue
        except ValueError as e:
            logging.getLogger("coretexpylib").warning(f">> [Coretex] Failed to override enum parameter \"{self.name}\". | {e}")
            return self.value



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Union, List
from pathlib import Path

from ..utils import command
from ..entities import CustomDataset


def cutadaptTrim(
    forwardFile: Union[str, Path],
    forwardOutput: Union[str, Path],
    forwardAdapter: str,
    reverseFile: Optional[Union[str, Path]] = None,
    reverseOutput: Optional[Union[str, Path]] = None,
    reverseAdapter: Optional[str] = None
) -> None:
    """
        Used to trim adapter sequences from sigle-end and paired-end sequencing reads

        Parameters
        ----------
        forwardFile : str
            Path to the file holding forward sequences
        forwardOutput : str
            Path to the output file for forward sequences
        forwardAdapter : str
            The adapter sequence for the forward reads
        reverseFile : Optional[str]
            Path to the file holding reverse sequences (pass for paired-end reads,
             otherwise only forward is required for single-end)
        reverseOutput : Optional[str]
            Path to the output file for reverse sequences (pass for paired-end reads,
             otherwise only forward is required for single-end)
        reverseAdapter : Optional[str]
            The adapter sequence for the reverse reads (pass for paired-end reads,
             otherwise only forward is required for single-end)
    """

    if isinstance(forwardFile, Path):
        forwardFile = str(forwardFile)

    if isinstance(forwardOutput, Path):
        forwardOutput = str(forwardOutput)

    if isinstance(reverseFile, Path):
        reverseFile = str(reverseFile)

    if isinstance(reverseOutput, Path):
        reverseOutput = str(reverseOutput)

    args: List[str] = [
        "cutadapt",
        "-o", forwardOutput,
        "-g", forwardAdapter,
    ]

    if reverseOutput is not None and reverseAdapter is not None:
        args.extend([
            "-p", reverseOutput,
            "-G", reverseAdapter
        ])

    args.append(forwardFile)
    if reverseFile is not None:
        args.append(reverseFile)

    command(args)


def isPairedEnd(dataset: CustomDataset) -> bool:
    """
        Check to see if the dataset has paired-end sequences, i.e. two fastq files
        per sample (excluding the metadata file)

        Parameters
        ----------
        dataset : CustomDataset
            Coretex dataset that will be checked for paired-end sequences

        Returns
        -------
        bool -> True if paired-end, False otherwise
    """

    for sample in dataset.samples:
        sample.unzip()

        if sample.name.startswith("_metadata"):
            continue

        if len(list(sample.path.glob("*.fastq*"))) != 2:
            return False

    return True



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TypeVar, Generic, Type, Generator, Optional, Union, Any
from pathlib import Path

import logging
import zipfile

from .dataset import Dataset
from ..sample import LocalSample, AnyLocalSample


SampleType = TypeVar("SampleType", bound = "LocalSample")
SampleGenerator = Generator[SampleType, None, None]


def _generateZippedSamples(path: Path, sampleClass: Type[SampleType]) -> Generator[SampleType, None, None]:
    for samplePath in path.glob("*"):
        if not zipfile.is_zipfile(samplePath):
            continue

        yield sampleClass(samplePath)


class LocalDataset(Generic[SampleType], Dataset[SampleType]):

    """
        Represents the generic Local Dataset class for all
        LocalDataset classes \n
        Used for working with local datasets

        Properties
        ----------
        path : Path
            local path of dataset
        sampleClass : Type[SampleType]
            class of sample
        generator : Optional[SampleGenerator]
            sample generator
    """

    def __init__(self, path: Path, sampleClass: Type[SampleType], generator: Optional[SampleGenerator] = None) -> None:
        if generator is None:
            generator = _generateZippedSamples(path, sampleClass)

        self.__path = path
        self.__sampleClass = sampleClass

        self.name = path.stem
        self.samples = list(generator)

    @staticmethod
    def default(path: Path) -> 'LocalDataset':
        """
            Creates Local Dataset object

            Parameters
            ----------
            path : Path
                Local Dataset path

            Returns
            -------
            LocalDataset -> Local Dataset object
        """

        return LocalDataset(path, LocalSample)

    @staticmethod
    def custom(path: Path, generator: SampleGenerator) -> 'LocalDataset':
        """
            Creates Custom Local Dataset object

            Parameters
            ----------
            path : Path
                Local Dataset path
            generator : SampleGenerator
                sample generator

            Returns
            -------
            LocalDataset -> Local Dataset object
        """

        return LocalDataset(path, AnyLocalSample, generator)

    @property
    def path(self) -> Path:
        """
            Returns
            -------
            Path -> Local Dataset path
        """

        return self.__path

    def download(self, decrypt: bool = True, ignoreCache: bool = False) -> None:
        logging.getLogger("coretexpylib").warning(">> [Coretex] Local dataset cannot be downloaded")

    def add(self, samplePath: Union[Path, str], sampleName: Optional[str] = None, **metadata: Any) -> SampleType:
        if isinstance(samplePath, str):
            samplePath = Path(samplePath)

        sample = self.__sampleClass(samplePath)
        self.samples.append(sample)

        return sample



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Any, Optional, Callable
from functools import wraps

import sys

from py3nvml import py3nvml

import click

from ...utils.process import command


def updateLib() -> None:
    command([sys.executable, "-m", "pip", "install", "--no-cache-dir", "--upgrade", "coretex"], ignoreStdout = True, ignoreStderr = True)


def isGPUAvailable() -> bool:
    try:
        py3nvml.nvmlInit()
        py3nvml.nvmlShutdown()
        return True
    except:
        return False


def onBeforeCommandExecute(fun: Callable[..., Any], excludeOptions: Optional[List[str]] = None) -> Any:
    if excludeOptions is None:
        excludeOptions = []

    def decorator(f: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(f)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            for key, value in click.get_current_context().params.items():
                if key in excludeOptions and value:
                    return f(*args, **kwargs)

            fun()
            return f(*args, **kwargs)
        return wrapper
    return decorator



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Generator
from pathlib import Path

import os

from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from .constants import DEFAULT_CHUNK_SIZE, AES_KEY_SIZE, IV_SIZE, AES_BLOCK_SIZE
from .utils import DataBuffer


class StreamEncryptor:

    """
        Implements functionality for encrypting a stream of bytes
        using AES 256.
    """

    def __init__(self, key: bytes, iv: Optional[bytes] = None, chunkSize: int = DEFAULT_CHUNK_SIZE) -> None:
        if len(key) != AES_KEY_SIZE:
            raise ValueError(f"AES key size: {len(key)}, but {AES_KEY_SIZE} expected")

        if iv is None:
            # generate random secure IV (initialization vector) if one wasn't provided
            iv = os.urandom(IV_SIZE)

        if len(iv) != IV_SIZE:
            raise ValueError(f"IV size: {len(iv)}, but {IV_SIZE} expected")

        self._key = key
        self._encryptor = Cipher(algorithms.AES256(self._key), modes.CBC(iv)).encryptor()
        self._buffer = DataBuffer()

        self.iv = iv
        self.chunkSize = chunkSize

    def feed(self, data: bytes) -> Generator[bytes, None, None]:
        """
            Encrypts the data using AES 256.

            Parameters
            ----------
            data : bytes
                data which should be encrypted

            Returns
            -------
            Generator[bytes, None, None] -> yields encrypted data in chunks
        """

        self._buffer.append(data)

        while self._buffer.remaining >= self.chunkSize:
            chunk = self._buffer.getBytes(self.chunkSize)
            yield self._encryptor.update(chunk)

    def flush(self) -> bytes:
        """
            Encrypts any remaining data using AES 256.

            Returns
            -------
            bytes -> encrypted data
        """

        chunk = self._buffer.getRemaining()

        # Pad the data if needed
        if len(chunk) % AES_BLOCK_SIZE != 0:
            padder = padding.PKCS7(AES_BLOCK_SIZE * 8).padder()
            chunk = padder.update(chunk) + padder.finalize()

        return self._encryptor.update(chunk) + self._encryptor.finalize()


def encryptFile(key: bytes, sourcePath: Path, destinationPath: Path) -> None:
    """
        Encrypts a file using AES 256. IV gets stored as the first 16 bytes
        of the encrypted file. IV is generated using "os.urandom".

        Parameters
        ----------
        key : bytes
            key which will be used for encrypting
        sourcePath : Path
            path to the file which will be encrypted
        destinationPath : Path
            path to the encrypted file

        Raises
        ------
        RuntimeError -> if file size % AES_BLOCK_SIZE has remainder
    """

    with sourcePath.open("rb") as source, destinationPath.open("wb") as destination:
        encryptor = StreamEncryptor(key)
        destination.write(encryptor.iv)

        while (chunk := source.read(DEFAULT_CHUNK_SIZE)):
            for encryptedData in encryptor.feed(chunk):
                destination.write(encryptedData)

        destination.write(encryptor.flush())

    if destinationPath.stat().st_size % AES_BLOCK_SIZE != 0:
        raise RuntimeError("File was corrupted during encryption")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Optional, Dict
from pathlib import Path

import json


CONFIG_PATH = Path("~/.config/coretex/config.json").expanduser()


# Only used by NetworkManager, should not be used anywhere else
class UserData:

    def __init__(self) -> None:
        with open(CONFIG_PATH, "r") as configFile:
            self.__values: Dict[str, Any] = json.load(configFile)

    def __getOptionalStr(self, key: str) -> Optional[str]:
        value = self.__values[key]
        if value is None:
            return None

        if isinstance(value, str):
            return value

        raise ValueError(f">> [Coretex] {key} is not of type optional str")

    def __setValue(self, key: str, value: Any) -> None:
        if not key in self.__values:
            raise KeyError(f">> [Coretex] {key} not found")

        self.__values[key] = value

        with open(CONFIG_PATH, "w") as configFile:
            json.dump(self.__values, configFile, indent = 4)

    @property
    def hasStoredCredentials(self) -> bool:
        return self.username is not None and self.password is not None

    @property
    def username(self) -> Optional[str]:
        return self.__getOptionalStr("username")

    @username.setter
    def username(self, value: Optional[str]) -> None:
        self.__setValue("username", value)

    @property
    def password(self) -> Optional[str]:
        return self.__getOptionalStr("password")

    @password.setter
    def password(self, value: Optional[str]) -> None:
        self.__setValue("password", value)

    @property
    def apiToken(self) -> Optional[str]:
        return self.__getOptionalStr("token")

    @apiToken.setter
    def apiToken(self, value: Optional[str]) -> None:
        self.__setValue("token", value)

    @property
    def refreshToken(self) -> Optional[str]:
        return self.__getOptionalStr("refreshToken")

    @refreshToken.setter
    def refreshToken(self, value: Optional[str]) -> None:
        self.__setValue("refreshToken", value)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

# from .token import Token
# from .transcription import Transcription
# from .utils import getTxtFilePath
# from ..coretex import CustomSample


# def loadTxtSample(sample: CustomSample) -> Transcription:
#     """
#         Tokenizes text sample

#         Parameters
#         ----------
#         sample : CustomSample
#         sample to be tokenized

#         Returns
#         -------
#         Transcription -> text and a list of tokens contained in the text

#         Raises
#         ------
#         ValueError -> if provided sample is not a valid text sample
#     """

#     path = getTxtFilePath(sample)
#     if path is None:
#         raise ValueError(f">> [Coretex] {sample.name} does not contain a valid txt file")

#     with path.open("r") as txtFile:
#         text = "\n".join(txtFile.readlines()).strip()

#     return Transcription.create(text, Token.fromText(text))



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, List, Dict, Optional, Union

from tabulate import tabulate

import click
import inquirer

from .node_mode import NodeMode
from .config_defaults import DEFAULT_CPU_COUNT


def clickPrompt(
    text: str,
    default: Any = None,
    type: Optional[Union[type, click.ParamType]] = None,
    **kwargs: Any
) -> Any:

    return click.prompt(click.style(f"\n\U00002754 {text}", fg = "cyan"), default = default, type = type, **kwargs)


def arrowPrompt(choices: List[Any], message: str) -> Any:
    click.echo("\n")
    answers = inquirer.prompt([
        inquirer.List(
            "option",
            message = message,
            choices = choices,
        )
    ])

    return answers["option"]


def previewConfig(config: Dict[str, Any]) -> None:
    allowDocker = "Yes" if config.get("allowDocker", False) else "No"

    if config.get("nodeSecret") is None or config.get("nodeSecret") == "":
        nodeSecret = ""
    else:
        nodeSecret = "********"

    table = [
        ["Node name",                   config["nodeName"]],
        ["Server URL",                  config["serverUrl"]],
        ["Coretex Node type",           config["image"]],
        ["Storage path",                config["storagePath"]],
        ["RAM",                         f"{config['nodeRam']}GB"],
        ["SWAP memory",                 f"{config['nodeSwap']}GB"],
        ["POSIX shared memory",         f"{config['nodeSharedMemory']}GB"],
        ["CPU cores allocated",         config.get("cpuCount", DEFAULT_CPU_COUNT)],
        ["Coretex Node mode",           f"{NodeMode(config['nodeMode']).name}"],
        ["Docker access",               allowDocker],
        ["Coretex Node secret",         nodeSecret],
        ["Coretex Node init script",    config.get("initScript", "")]
    ]

    if config.get("modelId") is not None:
        table.append(["Coretex Model ID", config["modelId"]])

    if config.get("nearWalletId") is not None:
        table.append(["NEAR wallet id", config["nearWalletId"]])

    stdEcho(tabulate(table))


def outputUrl(entityUrl: str) -> str:
    return ("\033[4m" + f"https://app.coretex.ai/{entityUrl}" + "\033[0m")


def stdEcho(text: str) -> None:
    click.echo(click.style(f"\n{text}", fg = "cyan"))


def successEcho(text: str) -> None:
    click.echo(click.style(f"\n\U0001F680 {text} \U0001F680", fg = "green"))


def progressEcho(text: str) -> None:
    click.echo(click.style(f"\n\U00002699 {text} \U00002699", fg = "yellow"))


def errorEcho(text: str) -> None:
    click.echo(click.style(f"\n\U0000274C {text} \U0000274C", fg = "red"))


def highlightEcho(text: str) -> None:
    click.echo(click.style(f"\n\U00002755 {text} \U00002755"))



from typing import Dict, Any, List, Tuple

import json

from .process import command, CommandException


def isDockerAvailable() -> None:
    try:
        # Run the command to check if Docker exists and is available
        command(["docker", "ps"], ignoreStdout = True, ignoreStderr = True)
    except CommandException:
        raise RuntimeError("Docker not available. Please check that it is properly installed and running on your system.")


def networkExists(name: str) -> bool:
    # This function inspects the specified Docker network using the
    # 'docker network inspect' command. If the command exits with a return code
    # of 0, indicating success, the function returns True, meaning the network exists.
    # If the command exits with a non-zero return code, indicating failure,
    # the function returns False, meaning the network doesn't exist.
    try:
        command(["docker", "network", "inspect", name], ignoreStdout = True, ignoreStderr = True)
        return True
    except:
        return False


def containerRunning(name: str) -> bool:
    try:
        _, output, _ = command(["docker", "ps", "--format", "{{.Names}}"], ignoreStderr = True, ignoreStdout = True)
        return name in output.splitlines()
    except:
        return False


def containerExists(name: str) -> bool:
    try:
        _, output, _ = command(["docker", "ps", "-a", "--format", "{{.Names}}"], ignoreStderr = True, ignoreStdout = True)
        return name in output.splitlines()
    except:
        return False


def createNetwork(name: str) -> None:
    if networkExists(name):
        removeNetwork(name)

    command(["docker", "network", "create", "--driver", "bridge", name], ignoreStdout = True)


def removeNetwork(name: str) -> None:
    command(["docker", "network", "rm", name], ignoreStdout = True, ignoreStderr = True)


def removeImage(image: str) -> None:
    command(["docker", "image", "rm", image], ignoreStdout = True, ignoreStderr = True)


def removeDanglingImages(repository: str, tag: str) -> None:
    _, output, _ = command(["docker", "image", "ls", repository, "--format", "json"], ignoreStdout = True, ignoreStderr = True)
    images = output.strip().split("\n")

    for image in images:
        if len(image) == 0:
            continue

        jsonImg = json.loads(image)
        if jsonImg["Tag"] != tag:
            removeImage(jsonImg["ID"])


def imagePull(image: str) -> None:
    command(["docker", "image", "pull", image])


def start(
    name: str,
    image: str,
    allowGpu: bool,
    ram: int,
    swap: int,
    shm: int,
    cpuCount: int,
    environ: Dict[str, str],
    volumes: List[Tuple[str, str]]
) -> None:

    runCommand = [
        "docker", "run", "-d",
        "--restart", 'always',
        "-p", "21000:21000",
        "--cap-add", "SYS_PTRACE",
        "--network", name,
        "--memory", f"{ram}G",
        "--memory-swap", f"{swap}G",
        "--shm-size", f"{shm}G",
        "--cpus", f"{cpuCount}",
        "--name", name,
    ]

    for key, value in environ.items():
        runCommand.extend(["--env", f"{key}={value}"])

    for source, destination in volumes:
        runCommand.extend(["-v", f"{source}:{destination}"])

    if allowGpu:
        runCommand.extend(["--gpus", "all"])

    # Docker image must always be the last parameter of docker run command
    runCommand.append(image)
    command(runCommand, ignoreStdout = True, ignoreStderr = True)


def stopContainer(name: str) -> None:
    command(["docker", "stop", name], ignoreStdout = True, ignoreStderr = True)


def removeContainer(name: str) -> None:
    command(["docker", "rm", name], ignoreStdout = True, ignoreStderr = True)


def manifestInspect(image: str) -> Dict[str, Any]:
    _, output, _ = command(["docker", "manifest", "inspect", image, "--verbose"], ignoreStdout = True)
    jsonOutput = json.loads(output)
    if not isinstance(jsonOutput, dict):
        raise TypeError(f"Invalid function result type \"{type(jsonOutput)}\". Expected: \"dict\"")

    return jsonOutput


def imageInspect(image: str) -> Dict[str, Any]:
    _, output, _ = command(["docker", "image", "inspect", image], ignoreStdout = True, ignoreStderr = True)
    jsonOutput = json.loads(output)
    if not isinstance(jsonOutput, list):
        raise TypeError(f"Invalid json.loads() result type \"{type(jsonOutput)}\". Expected: \"list\"")

    if not isinstance(jsonOutput[0], dict):  # Since we are inspecting image with specific repository AND tag output will be a list with single object
        raise TypeError(f"Invalid function result type \"{type(jsonOutput[0])}\". Expected: \"dict\"")

    return jsonOutput[0]


def getResourceLimits() -> Tuple[int, int]:
    _, output, _ = command(["docker", "info", "--format", "{{json .}}"], ignoreStdout = True, ignoreStderr = True)
    jsonOutput = json.loads(output)

    return jsonOutput["NCPU"], round(jsonOutput["MemTotal"] / (1024 ** 3))



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TypeVar, Generic, Dict, Any, List
from typing_extensions import override
from datetime import datetime
from pathlib import Path

import os
import time
import shutil

from .sample import Sample
from ..project import ProjectType
from ... import folder_manager
from ...codable import KeyDescriptor
from ...networking import NetworkObject, networkManager, NetworkRequestError, \
    fileChunkUpload, MAX_CHUNK_SIZE, FileData
from ...utils import TIME_ZONE
from ...cryptography import getProjectKey, aes


SampleDataType = TypeVar("SampleDataType")


def _relinkSample(samplePath: Path) -> None:
    for datasetPath in folder_manager.datasetsFolder.iterdir():
        linkPath = datasetPath / samplePath.name
        if not linkPath.exists():
            continue

        linkPath.unlink()
        os.link(samplePath, linkPath)


class NetworkSample(Generic[SampleDataType], Sample[SampleDataType], NetworkObject):

    """
        Represents a base class for all Sample classes which are
        comunicating with Coretex.ai
    """

    isLocked: bool
    projectId: int
    projectType: ProjectType
    lastModified: datetime
    isEncrypted: bool

    @property
    def path(self) -> Path:
        """
            Returns
            -------
            Path -> path for network sample
        """

        return folder_manager.samplesFolder / str(self.id)

    @property
    def zipPath(self) -> Path:
        """
            Returns
            -------
            Path -> zip path for network sample
        """

        return self.path.with_suffix(".zip")

    @property
    def downloadPath(self) -> Path:
        """
            Returns
            -------
            Path -> path to which the network sample is downloaded to
        """

        return self.path.with_suffix(".bin") if self.isEncrypted else self.zipPath

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()

        descriptors["projectType"] = KeyDescriptor("project_task", ProjectType)
        descriptors["lastModified"] = KeyDescriptor("storage_last_modified", datetime)

        return descriptors

    @classmethod
    def _endpoint(cls) -> str:
        return "session"

    def modifiedSinceLastDownload(self) -> bool:
        """
            Checking if sample has been modified since last download, if the sample is already
            stored locally

            Returns
            -------
            bool -> False if sample has not changed since last download, True otherwise

            Raises
            ------
            FileNotFoundError -> sample file cannot be found
        """

        if not self.downloadPath.exists():
            raise FileNotFoundError(
                f">> [Coretex] Sample file could not be found at {self.downloadPath}. "
                "Cannot check if file has been modified since last download"
            )

        lastModified = datetime.fromtimestamp(self.downloadPath.stat().st_mtime).astimezone(TIME_ZONE)
        return self.lastModified > lastModified

    def decrypt(self, ignoreCache: bool = False) -> None:
        """
            Decrypts the content of this Sample and caches
            the results. Is ignored if the "isEncrypted" value is False.

            Parameters
            ----------
            ignoreCache : bool
                defines if content should be decrypted if a cache for decryption
                already exists
        """

        if not self.isEncrypted:
            return

        if ignoreCache and self.zipPath.exists():
            self.zipPath.unlink()

        if not ignoreCache and self.zipPath.exists():
            return

        # Decrypt sample
        aes.decryptFile(getProjectKey(self.projectId), self.downloadPath, self.zipPath)

        # Relink sample to all datasets to which it belongs
        _relinkSample(self.zipPath)

    def _download(self, ignoreCache: bool = False) -> None:
        if self.downloadPath.exists() and self.modifiedSinceLastDownload():
            ignoreCache = True

        if ignoreCache and self.downloadPath.exists():
            self.downloadPath.unlink()
            if self.path.exists():
                shutil.rmtree(self.path)

            if self.zipPath.exists() and self.isEncrypted:
                self.zipPath.unlink()

        if not ignoreCache and self.downloadPath.exists():
            return

        params = {
            "id": self.id
        }

        response = networkManager.streamDownload(
            f"{self._endpoint()}/export",
            self.downloadPath,
            params
        )

        if response.hasFailed():
            raise NetworkRequestError(response, f"Failed to download Sample \"{self.name}\"")

    @override
    def download(self, decrypt: bool = True, ignoreCache: bool = False) -> None:
        """
            Downloads and optionally decrypts sample from Coretex.ai

            Raises
            ------
            NetworkRequestError -> if some kind of error happened during
            the download process
        """

        if decrypt and not self.isEncrypted:
            # Change to false if sample is not encrypted
            decrypt = False

        # Download the sample
        self._download(ignoreCache)

        if decrypt:
            # Decrypt the sample
            self.decrypt(ignoreCache)

        # Update sample download time to now
        os.utime(self.downloadPath, (os.stat(self.downloadPath).st_atime, time.time()))

        # If sample was downloaded succesfully relink it to datasets to which it is linked
        _relinkSample(self.downloadPath)

    @override
    def unzip(self, ignoreCache: bool = False) -> None:
        if not self.downloadPath.exists():
            raise RuntimeError("You must first download the Sample before you can unzip it")

        if not self.zipPath.exists() and self.isEncrypted:
            raise RuntimeError("You must first decrypt the Sample before you can unzip it")

        super().unzip(ignoreCache)

    @override
    def load(self) -> SampleDataType:
        return super().load()  # type: ignore

    @override
    def _updateArchive(self) -> None:
        super()._updateArchive()

        if self.isEncrypted:
            aes.encryptFile(getProjectKey(self.projectId), self.zipPath, self.downloadPath)

    def _overwriteSample(self, samplePath: Path) -> None:
        if not self.isEncrypted:
            raise RuntimeError("Only encrypted samples can be overwriten.")

        with folder_manager.tempFile() as encryptedPath:
            aes.encryptFile(getProjectKey(self.projectId), samplePath, encryptedPath)

            params: Dict[str, Any] = {
                "id": self.id
            }

            files: List[FileData] = []

            # Use chunk upload if file is larger than MAX_CHUNK_SIZE
            # Use normal upload if file is smaller than MAX_CHUNK_SIZE
            size = encryptedPath.stat().st_size

            if size > MAX_CHUNK_SIZE:
                params["file_id"] = fileChunkUpload(encryptedPath)
            else:
                files.append(FileData.createFromPath("file", encryptedPath))

            response = networkManager.formData(f"{self._endpoint()}/upload", params, files, timeout = (5, 300))
            if response.hasFailed():
                raise NetworkRequestError(response, f"Failed to overwrite Sample \"{self.name}\"")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple, Dict, List

import time
import logging

from ...entities import MetricType, TaskRun, Metric
from ...entities.task_run.metrics import metric_factory


def _getMetrics() -> List[Metric]:
    metrics = [
        metric_factory.createMetric("cpu_usage", "time (s)", MetricType.interval, "usage (%)", MetricType.percent, None, [0, 100]),
        metric_factory.createMetric("ram_usage", "time (s)", MetricType.interval, "usage (%)", MetricType.percent, None, [0, 100]),
        metric_factory.createMetric("swap_usage", "time (s)", MetricType.interval, "usage (%)", MetricType.percent, None, [0, 100]),
        metric_factory.createMetric("download_speed", "time (s)", MetricType.interval, "bytes", MetricType.bytes),
        metric_factory.createMetric("upload_speed", "time (s)", MetricType.interval, "bytes", MetricType.bytes),
        metric_factory.createMetric("disk_read", "time (s)", MetricType.interval, "bytes", MetricType.bytes),
        metric_factory.createMetric("disk_write", "time (s)", MetricType.interval, "bytes", MetricType.bytes)
    ]

    # If GPU exists add GPU related metrics to the list
    try:
        from py3nvml import py3nvml

        # Do not shutdown otherwise when extracting gpu metrics it will throw error
        py3nvml.nvmlInit()

        metrics.extend([
            metric_factory.createMetric("gpu_usage", "time (s)", MetricType.interval, "usage (%)", MetricType.percent, None, [0, 100]),
            metric_factory.createMetric("gpu_temperature", "time (s)", MetricType.interval, "usage (%)", MetricType.percent)
        ])

        logging.getLogger("coretexpylib").debug(">> [Coretex] Initialized GPU metrics")
    except:
        logging.getLogger("coretexpylib").debug(">> [Coretex] Failed to initialize GPU metrics")

    return metrics


def create(taskRun: TaskRun) -> None:
    taskRun.createMetrics(_getMetrics())


def upload(taskRun: TaskRun) -> None:
    x = time.time()
    metricValues: Dict[str, Tuple[float, float]] = {}

    for metric in taskRun.metrics:
        metricValue = metric.extract()

        if metricValue is not None:
            metricValues[metric.name] = x, metricValue

    taskRun.submitMetrics(metricValues)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Optional, Tuple
from pathlib import Path

import logging
import subprocess


def logProcessOutput(output: bytes, level: int) -> None:
    decoded = output.decode("UTF-8")

    for line in decoded.split("\n"):
        # skip empty lines
        if line.strip() == "":
            continue

        # ignoring type for now, has to be fixed in coretexpylib
        logging.getLogger("coretexpylib").log(level, line)


class CommandException(Exception):
    pass


def command(
    args: List[str],
    ignoreStdout: bool = False,
    ignoreStderr: bool = False,
    shell: bool = False,
    check: bool = True
) -> Tuple[int, str, str]:

    process = subprocess.Popen(
        args,
        shell = shell,
        cwd = Path.cwd(),
        stdout = subprocess.PIPE,
        stderr = subprocess.PIPE
    )

    stdOutStr = ""
    stdErrStr = ""

    returnCode: Optional[int] = None

    stdout = process.stdout
    stderr = process.stderr

    if stdout is None and not ignoreStdout:
        commandArgs = " ".join(args)
        raise ValueError(f">> [Coretex] Something went wrong while trying to execute \"{commandArgs}\"")

    while (returnCode := process.poll()) is None:
        if stdout is not None:
            line = stdout.readline()
            stdOutStr = "\n".join([stdOutStr, line.decode("utf-8")])
            if not ignoreStdout:
                logProcessOutput(line, logging.INFO)

    if stderr is None and not ignoreStderr:
        commandArgs = " ".join(args)
        raise ValueError(f">> [Coretex] Something went wrong while trying to execute \"{commandArgs}\"")

    if stderr is not None:
        lines = stderr.readlines()
        stdErrStr = b"".join(lines).decode("utf-8")
        if not ignoreStderr:
            logProcessOutput(b"".join(lines), logging.WARNING if returnCode == 0 else logging.FATAL)

    if returnCode != 0 and check:
        commandArgs = " ".join(args)
        raise CommandException(f">> [Coretex] Failed to execute command \"{commandArgs}\". Exit code \"{returnCode}\"")

    return returnCode, stdOutStr, stdErrStr



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from base64 import b64decode

import os


def _projectKeyEnvName(projectId: int) -> str:
    return f"CTX_PROJECT_KEY_{projectId}"


def getProjectKey(projectId: int) -> bytes:
    """
        Retrieves encryption key for provided project if
        the executing machine has been authorized to access it

        Parameters
        ----------
        projectId : int
            project whose key is being retrieved

        Returns
        -------
        bytes -> encryption key
    """

    envName = _projectKeyEnvName(projectId)
    if envName not in os.environ:
        raise RuntimeError(f"Not authorized to access project: {projectId}")

    encodedKey = os.environ[envName]
    return b64decode(encodedKey)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime, timezone


# Time zone used for dates on Coretex backend
TIME_ZONE = timezone.utc

# Date format used by Coretex backend
DATE_FORMAT = "%Y-%m-%d %H:%M:%S.%f%z"

CORETEX_DATE_FORMATS = [
    DATE_FORMAT,
    "%Y-%m-%dT%H:%M:%S.%f%z",
    "%Y-%m-%d %H:%M:%S%z"
]


def decodeDate(value: str) -> datetime:
    """
        Converts the date to a format used by Coretex

        Returns
        -------
        datetime -> object whose datetime is represented using the Coretex datetime format
    """

    for format in CORETEX_DATE_FORMATS:
        try:
            return datetime.strptime(value, format)
        except ValueError:
            continue

    for format in CORETEX_DATE_FORMATS:
        try:
            # Python's datetime library requires UTC minutes to always
            # be present in the date in either of those 2 formats:
            # - +HHMM
            # - +HH:MM
            # BUT coretex API sends it in one of those formats:
            # - +HH
            # - +HH:MM (only if the minutes have actual value)
            # so we need to handle the first case where minutes
            # are not present by adding them manually
            return datetime.strptime(f"{value}00", format)
        except ValueError:
            continue

    raise ValueError(f"Failed to convert \"{value}\" to any of the supported formats \"{CORETEX_DATE_FORMATS}\"")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

# from typing import List, Optional
# from typing_extensions import Self

# from deepspeech import TokenMetadata

# from ..codable import Codable


# class Token(Codable):

#     """
#         Represents a single token from text

#         Properties
#         ----------
#         text : str
#             textual value of the token
#         startIndex : int
#             starting index of the token
#         endIndex : int
#             ending index of the token
#         startTime : Optional[float]
#             starting time of token in audio file
#             (only if token was extracted from audio transcription)
#         endTime : Optional[float]
#             ending time of token in audio file
#             (only if token was extracted from audio transcription)
#     """

#     text: str
#     startIndex: int
#     endIndex: int
#     startTime: Optional[float]
#     endTime: Optional[float]

#     @classmethod
#     def create(
#         cls,
#         text: str,
#         startIndex: int,
#         endIndex: int,
#         startTime: Optional[float],
#         endTime: Optional[float]
#     ) -> Self:

#         obj = cls()

#         obj.text = text
#         obj.startIndex = startIndex
#         obj.endIndex = endIndex
#         obj.startTime = startTime
#         obj.endTime = endTime

#         return obj

#     @classmethod
#     def fromTokenMetadata(cls, tokenMetadata: List[TokenMetadata]) -> List[Self]:
#         """
#             Creates a list of tokens from output of the deepspeech model

#             Parameters
#             ----------
#             tokenMetadata : List[TokenMetadata]
#             output of deepspeech model

#             Returns
#             -------
#             List[Self] -> list of tokens
#         """

#         tokens: List[Self] = []

#         startIndex: Optional[int] = None
#         startTime: Optional[float] = None
#         characters: List[str] = []

#         for currentIndex, element in enumerate(tokenMetadata):
#             if startIndex is None and len(characters) == 0:
#                 startIndex = currentIndex

#             if startTime is None and len(characters) == 0:
#                 startTime = element.start_time

#             if element.text.isspace() and startIndex is not None and startTime is not None and len(characters) > 0:
#                 token = cls.create("".join(characters), startIndex, currentIndex, startTime, element.start_time)
#                 tokens.append(token)

#                 startIndex = None
#                 startTime = None
#                 characters.clear()

#                 continue

#             if not element.text.isspace():
#                 characters.append(element.text)

#         if startIndex is not None and startTime is not None and len(characters) > 0:
#             token = cls.create("".join(characters), startIndex, currentIndex, startTime, element.start_time)
#             tokens.append(token)

#         return tokens

#     @classmethod
#     def fromText(cls, text: str) -> List[Self]:
#         """
#             Tokenizes provided text

#             Parameters:
#             text: str -> text to be tokenized

#             Retursn:
#             List[Self] -> list of tokens
#         """

#         tokens: List[Self] = []

#         startIndex: Optional[int] = None
#         characters: List[str] = []

#         for currentIndex, character in enumerate(text):
#             if startIndex is None and len(characters) == 0:
#                 startIndex = currentIndex

#             if character.isspace() and startIndex is not None and len(characters) > 0:
#                 token = cls.create("".join(characters), startIndex, currentIndex, None, None)
#                 tokens.append(token)

#                 startIndex = None
#                 characters.clear()

#                 continue

#             if not character.isspace():
#                 characters.append(character)

#         if startIndex is not None and len(characters) > 0:
#             token = cls.create("".join(characters), startIndex, currentIndex, None, None)
#             tokens.append(token)

#         return tokens



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union
from pathlib import Path

import logging

from .network_manager_base import FileData
from .network_manager import networkManager
from .network_response import NetworkRequestError


MAX_CHUNK_SIZE = 128 * 1024 * 1024  # 128 MiB


def _loadChunk(filePath: Path, start: int, chunkSize: int) -> bytes:
    with filePath.open("rb") as file:
        file.seek(start)
        return file.read(chunkSize)


class ChunkUploadSession:

    """
        A class which splits a file into chunks and uploades it
        chunk by chunk. This class should be used for uploading
        files larger than 2 GiB, since Python does not support
        uploading files with a larger size.

        Maximum chunk size is 128 MiB.

        Properties
        ----------
        chunkSize : int
            size of chunks into which the file will be split
            maximum value is 128 MiB, while the minimum value is 1
        filePath : Union[Path, str]
            path to the file which will be uploaded
        fileSize : int
            size of the file which will be uploaded
    """

    def __init__(self, chunkSize: int, filePath: Union[Path, str]) -> None:
        if chunkSize <= 0 or chunkSize > MAX_CHUNK_SIZE:
            raise ValueError(f">> [Coretex] Invalid \"chunkSize\" value \"{chunkSize}\". Value must be in range 0-{MAX_CHUNK_SIZE}")

        if isinstance(filePath, str):
            filePath = Path(filePath)

        self.chunkSize = chunkSize
        self.filePath = filePath
        self.fileSize = filePath.lstat().st_size

    def __start(self) -> str:
        parameters = {
            "size": self.fileSize
        }

        response = networkManager.post("upload/start", parameters)
        if response.hasFailed():
            raise NetworkRequestError(response, f"Failed to start chunked upload for \"{self.filePath}\"")

        uploadId = response.getJson(dict).get("id")

        if not isinstance(uploadId, str):
            raise ValueError(f">> [Coretex] Invalid API response, invalid value \"{uploadId}\" for field \"id\"")

        return uploadId

    def __uploadChunk(self, uploadId: str, start: int, end: int) -> None:
        parameters = {
            "id": uploadId,
            "start": start,
            "end": end - 1  # API expects start/end to be inclusive
        }

        chunk = _loadChunk(self.filePath, start, self.chunkSize)
        files = [
            FileData.createFromBytes("file", chunk, self.filePath.name)
        ]

        response = networkManager.formData("upload/chunk", parameters, files)
        if response.hasFailed():
            raise NetworkRequestError(response, f"Failed to upload file chunk with byte range \"{start}-{end}\"")

        logging.getLogger("coretexpylib").debug(f">> [Coretex] Uploaded chunk with range \"{start}-{end}\"")

    def run(self) -> str:
        """
            Uploads the file to Coretex.ai

            Returns
            -------
            str -> ID of the uploaded file

            Raises
            ------
            NetworkRequestError, ValueError -> if some kind of error happened during
            the upload of the provided file

            Example
            -------
            >>> from coretex.networking import ChunkUploadSession, NetworkRequestError
            \b
            >>> chunkSize = 16 * 1024 * 1024  # chunk size: 16 MiB
            >>> uploadSession = ChunkUploadSession(chunkSize, path/fo/file.ext)
            \b
            >>> try:
                    uploadId = uploadSession.run()
                    print(uploadId)
                except NetworkRequestError, ValueError:
                    print("Failed to upload file")
        """
        logging.getLogger("coretexpylib").debug(f">> [Coretex] Starting upload for \"{self.filePath}\"")

        uploadId = self.__start()

        chunkCount = self.fileSize // self.chunkSize
        if self.fileSize % self.chunkSize != 0:
            chunkCount += 1

        for i in range(chunkCount):
            start = i * self.chunkSize
            end = min(start + self.chunkSize, self.fileSize)

            self.__uploadChunk(uploadId, start, end)

        return uploadId


def fileChunkUpload(path: Path, chunkSize: int = MAX_CHUNK_SIZE) -> str:
    """
        Uploads file in chunks to Coretex.ai server.
        Should be used when uploading large files.

        Parameters
        ----------
        path : Path
            File which will be uploaded in chunks
        chunkSize : int
            Size of the chunks into which file will be split
            before uploading. Maximum value is 128 MiBs

        Returns
        -------
        str -> id of the file which was uploaded
    """

    if not path.is_file():
        raise ValueError(f"{path} is not a file")

    if chunkSize > MAX_CHUNK_SIZE:
        chunkSize = MAX_CHUNK_SIZE

    uploadSession = ChunkUploadSession(MAX_CHUNK_SIZE, path)
    return uploadSession.run()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum


class ParameterType(Enum):

    # Single value parameters
    integer           = "int"
    floatingPoint     = "float"
    string            = "str"
    boolean           = "bool"
    dataset           = "dataset"
    model             = "model"
    imuVectors        = "IMUVectors"
    enum              = "enum"
    range             = "range"
    awsSecret         = "aws_secret"
    gitSecret         = "git_secret"
    credentialsSecret = "credentials_secret"

    # List parameters
    intList               = "list[int]"
    floatList             = "list[float]"
    strList               = "list[str]"
    datasetList           = "list[dataset]"
    modelList             = "list[model]"
    enumList              = "list[enum]"
    awsSecretList         = "list[aws_secret]"
    gitSecretList         = "list[git_secret]"
    credentialsSecretList = "list[credentials_secret]"



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Iterator, Optional
from pathlib import Path
from contextlib import contextmanager

import os
import shutil
import uuid

from .utils import file as file_utils


"""
    Used for handling everything related to local storage
    when working with Coretex

    Contains
    --------
    samplesFolder : str
        folder where samples are stored
    modelsFolder : str
        folder where models are stored
    temp : str
        folder where temp files and folders are stored,
        this is deleted when the run has finished executing

    datasetsFolder : Path
        folder where datasets are stored (samples are symlinked for datasets)
    cache : Path
        folder where cache module stores items
    logs : Path
        folder where node and run logs are stored
    environments : Path
        folder where node stores python environments
"""


def _createFolder(name: str) -> Path:
    path = _root / name

    if not path.exists():
        path.mkdir(parents = True, exist_ok = True)

    return path


_root = Path(os.environ["CTX_STORAGE_PATH"]).expanduser()

samplesFolder    = _createFolder("samples")
modelsFolder     = _createFolder("models")
datasetsFolder   = _createFolder("datasets")
cache            = _createFolder("cache")
logs             = _createFolder("logs")
environments     = _createFolder("environments")
temp             = _createFolder("temp")
_artifactsFolder = _createFolder("artifacts")

runsLogDirectory = logs / "runs"
runsLogDirectory.mkdir(exist_ok = True)

coretexpylibLogs = logs / "coretexpylib"
coretexpylibLogs.mkdir(exist_ok = True)


def createTempFolder(name: str) -> Path:
    """
        Creates temp folder which is deleted once
        the run has finished executing

        Parameters
        ----------
        name : str
            name of the folder

        Returns
        -------
        Path -> path to the created folder

        Raises
        ------
        FileExistsError -> if the temp folder already exists

        Example
        -------
        >>> from coretex import folder_manager
        \b
        >>> dummyFolderPath = folder_manager.createTempFolder("dummyTempFolder")
        >>> print(dummyFolderPath)
        "/Users/X/.coretex/temp/dummyTempFolder"
    """

    tempFolderPath = temp / name

    if tempFolderPath.exists():
        raise FileExistsError

    tempFolderPath.mkdir()
    return tempFolderPath


def getArtifactsFolder(taskRunId: int) -> Path:
    """
        Retrieves the path to where the artifacts are stored
        for the specified TaskRuns

        Parameters
        ----------
        taskRunId : int
            id of the TaskRun

        Returns
        -------
        Path -> path to the TaskRun artifacts local storage

        Example
        -------
        >>> from coretex.folder_management import FolderManager
        \b
        >>> artifactsFolderPath = FolderManager.instance().getArtifactsFolder(1023)
        >>> print(artifactsFolderPath)
        Path("/Users/bogdanbm/.coretex/artifacts/1023")

    """

    return _artifactsFolder / str(taskRunId)


def clearDirectory(path: Path) -> None:
    for element in file_utils.walk(path):
        if element.is_file():
            element.unlink()

        if element.is_dir():
            shutil.rmtree(element)


def clearTempFiles() -> None:
    """
        Deletes all temp files and folders (including artifacts)
    """

    clearDirectory(temp)
    clearDirectory(_artifactsFolder)


def getRunLogsDir(taskRunId: int) -> Path:
    taskRunLogsDir = runsLogDirectory / str(taskRunId)
    taskRunLogsDir.mkdir(parents = True, exist_ok = True)

    return taskRunLogsDir


@contextmanager
def tempFile(name: Optional[str] = None) -> Iterator[Path]:
    """
        Returns a path to temporary file and deletes
        it if it exists once the context is exited.

        Parameters
        ----------
        name : Optional[str]
            Name of the file. If not specified a random uuid4
            will be generated and used as the name

        Returns
        -------
        Iterator[Path] -> path to the file
    """

    if name is None:
        name = str(uuid.uuid4())

    path = temp / name
    if path.exists():
        raise FileExistsError(path)

    try:
        yield path
    finally:
        path.unlink(missing_ok = True)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Callable, List

import logging
import sys

from .remote import processRemote
from .current_task_run import setCurrentTaskRun
from .. import folder_manager
from ..entities import TaskRun, TaskRunStatus
from ..logging import createFormatter, initializeLogger
from ..logging.severity import LogSeverity
from ..networking import RequestFailedError


def _initializeLogger(taskRun: TaskRun) -> None:
    # enable/disable verbose mode for taskRuns
    severity = LogSeverity.info
    verbose = taskRun.parameters.get("verbose", False)

    if verbose:
        severity = LogSeverity.debug

    streamHandler = logging.StreamHandler(sys.stdout)
    streamHandler.setLevel(severity.getLevel())
    streamHandler.setFormatter(createFormatter(
        includeTime = False,
        includeLevel = False,
        includeColor = False,
        jsonOutput = True
    ))

    logPath = folder_manager.getRunLogsDir(taskRun.id) / "run.log"
    initializeLogger(severity, logPath, streamHandler)


def _prepareForExecution(taskRun: TaskRun) -> None:
    _initializeLogger(taskRun)

    taskRun.updateStatus(
        status = TaskRunStatus.inProgress,
        message = "Executing task."
    )


def initializeRTask(mainFunction: Callable[[TaskRun], None], args: List[str]) -> None:
    """
        Initializes and starts the R task as Coretex TaskRun

        Parameters
        ----------
        mainFunction : Callable[[ExecutingTaskRun], None]
            entry point function
        args : Optional[List[str]]
            list of command line arguments
    """

    taskRun, callback = processRemote(args)

    try:
        _prepareForExecution(taskRun)
        setCurrentTaskRun(taskRun)

        callback.onStart()

        logging.getLogger("coretexpylib").info("TaskRun execution started")
        mainFunction(taskRun)

        callback.onSuccess()
    except RequestFailedError:
        callback.onNetworkConnectionLost()
    except KeyboardInterrupt:
        callback.onKeyboardInterrupt()
    except BaseException as ex:
        callback.onException(ex)

        # sys.exit is ok here, finally block is guaranteed to execute
        # due to how sys.exit is implemented (it internally raises SystemExit exception)
        sys.exit(1)
    finally:
        callback.onCleanUp()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TextIO

import sys
import logging
import faulthandler
import signal

from .current_task_run import setCurrentTaskRun
from .. import folder_manager
from ..entities import TaskRun


class TaskCallback:

    def __init__(self, taskRun: TaskRun, outputStream: TextIO = sys.stdout) -> None:
        self._taskRun = taskRun

        # Store sys streams so they can be restored later
        self.__stdoutBackup = sys.stdout
        self.__stderrBackup = sys.stderr

        # Set sys output to new stream
        sys.stdout = outputStream
        sys.stderr = outputStream

    def _restoreStreams(self) -> None:
        sys.stdout = self.__stdoutBackup
        sys.stderr = self.__stderrBackup

    def onStart(self) -> None:
        # Call "kill -30 task_run_process_id" to dump current stack trace of the TaskRun into the file
        # 30 == signal.SIGUSR1
        # Only works on *nix systems

        faultHandlerPath = folder_manager.getRunLogsDir(self._taskRun.id) / "stacktrace.log"
        faulthandler.register(
            signal.SIGUSR1,
            file = faultHandlerPath.open("w"),
            all_threads = True
        )

    def onSuccess(self) -> None:
        logging.getLogger("coretexpylib").info(">> [Coretex] TaskRun finished successfully")

        self._restoreStreams()

    def onException(self, exception: BaseException) -> None:
        logging.getLogger("coretexpylib").critical(">> [Coretex] TaskRun failed to finish due to an error")
        logging.getLogger("coretexpylib").debug(exception, exc_info = True)
        logging.getLogger("coretexpylib").critical(str(exception))

        self._restoreStreams()

    def onKeyboardInterrupt(self) -> None:
        pass

    def onNetworkConnectionLost(self) -> None:
        sys.exit(1)

    def onCleanUp(self) -> None:
        # Flushes the internal buffers of logging module handlers
        # and other logging cleanup
        # IMPORTANT: do not use logging after calling this
        logging.shutdown()

        try:
            from py3nvml import py3nvml
            py3nvml.nvmlShutdown()
        except:
            pass

        setCurrentTaskRun(None)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import IntEnum


class TaskRunStatus(IntEnum):

    """
        List of possible TaskRun statuses during the TaskRun lifetime
    """

    queued = 1
    preparingToStart = 2
    inProgress = 3
    completedWithSuccess = 4
    completedWithError = 5
    stopped = 6
    stopping = 7
    startRequested = 8
    stopRequested = 9

    @property
    def defaultMessage(self) -> str:
        """
            List of supported statuses:
            - preparingToStart : TaskRun preparing to start
            - completedWithSuccess : TaskRun is completed without errors
            - completedWithError : TaskRun is completed with error
            - stopped : TaskRun is stopped manually
            - stopping : TaskRun is stopping

            Returns
            -------
            str -> Appropriate message based on TaskRun status

            Raises
            ------
            ValueError -> if unsupported status is provided
        """

        if self == TaskRunStatus.preparingToStart:
            return "Preparing to start the Task Run."

        if self == TaskRunStatus.completedWithSuccess:
            return "Task Run completed successfully."

        if self == TaskRunStatus.completedWithError:
            return "Task Run failed due to an error. View console output for more details."

        if self == TaskRunStatus.stopped:
            return "User stopped the Task Run."

        if self == TaskRunStatus.stopping:
            return "Stopping the Task Run."

        raise ValueError(f">> [Coretex] {self.name} has no default message")

    @property
    def isFinal(self) -> bool:
        """
            List of final statuses:
            - TaskRunStatus.completedWithSuccess : TaskRun finished without error
            - TaskRunStatus.completedWithError : TaskRun finished with an error
            - TaskRunStatus.stopped : TaskRun is manually stopped

            Returns
            -------
            bool -> True if a status is a final status for a run
        """

        return (
                self == TaskRunStatus.completedWithSuccess or
                self == TaskRunStatus.completedWithError or
                self == TaskRunStatus.stopped
        )



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Any
from pathlib import Path

import json

from .image_sample_data import AnnotatedImageSampleData
from .image_format import ImageFormat
from ..local_sample import LocalSample
from ...annotation import CoretexImageAnnotation


class LocalImageSample(LocalSample[AnnotatedImageSampleData]):

    """
        Represents the local Image Sample object\n
        Contains basic properties and functionality for local image Sample classes\n
        The class has several methods that allow users to access and
        manipulate local image data and annotations
    """

    @property
    def imagePath(self) -> Path:
        for format in ImageFormat:
            imagePaths = list(self.path.glob(f"*.{format.extension}"))
            imagePaths = [path for path in imagePaths if not "thumbnail" in str(path)]

            if len(imagePaths) > 0:
                return Path(imagePaths[0])

        raise FileNotFoundError

    @property
    def annotationPath(self) -> Path:
        return self.path / "annotations.json"

    @property
    def metadataPath(self) -> Path:
        return self.path / "metadata.json"

    def load(self) -> AnnotatedImageSampleData:
        """
            Loads image and its annotation if it exists

            Returns
            -------
            AnnotatedImageSampleData -> image data and annotation in Coretex.ai format
        """

        return AnnotatedImageSampleData(self.path)

    def loadMetadata(self) -> Dict[str, Any]:
        """
            Loads sample metadata into a dictionary

            Returns
            -------
            dict[str, Any] -> the metadata as a dict object

            Raises
            ------
            FileNotFoundError -> if metadata file is missing\n
            ValueError -> if json in the metadata file is list
        """

        if not self.metadataPath.exists():
            raise FileNotFoundError(f"Metadata file \"{self.metadataPath}\" not found")

        with self.metadataPath.open("r") as metadataFile:
            metadata = json.load(metadataFile)
            if not isinstance(metadata, dict):
                raise ValueError(f"Metatada for sample \"{self.name}\" is a list. Expected dictionary")

            return metadata

    def saveAnnotation(self, coretexAnnotation: CoretexImageAnnotation) -> bool:
        """
            Updates annotation for the image

            Returns
            -------
            bool -> returns True if successful, False otherwise
        """

        with self.annotationPath.open("w") as file:
            json.dump(coretexAnnotation.encode(), file)

        self._updateArchive()
        return True

    def saveMetadata(self, metadata: Dict[str, Any]) -> None:
        """
            Saves a json object as metadata for the sample

            Parameters
            ----------
            metadata : dict[str, Any]
                Json object containing sample metadata
        """

        with self.metadataPath.open("w") as file:
            json.dump(metadata, file)

        self._updateArchive()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Any, Dict, List
from typing_extensions import Self

import inflection

from .network_manager import networkManager
from .network_response import NetworkRequestError
from ..codable import Codable


DEFAULT_PAGE_SIZE = 100


class NetworkObject(Codable):

    """
        Base class for every Coretex.ai entity representation in Python

        Properties
        ----------
        id : int
            id of object
        isDeleted : bool
            boolean value that represents if object will be shown or not
    """

    id: int
    isDeleted: bool

    # Required init
    def __init__(self) -> None:
        pass

    @classmethod
    def _endpoint(cls) -> str:
        """
            Maps the entity endpoint to overriden value, else
            it uses inflection.underscode on the class name for endpoint

            Returns
            -------
            str -> Coretex.ai object endpoint for a given class
        """

        return inflection.underscore(cls.__name__)

    def entityUrl(self) -> str:
        """
            Maps the entity url to the frontend page url

            Returns
            -------
            str -> The URL path on Coretex for the entity.
        """

        return f"{type(self).__name__.lower()}/{self.id}"

    def __eq__(self, __o: object) -> bool:
        """
            Checks if the NetworkObjects which have id property
            defined are equal

            Parameter
            ---------
            __o : object
                object to which we are comparing self

            Returns
            -------
            bool -> True if ids are present and equal, False in any other case
        """

        # check if object parent class matches
        if isinstance(__o, NetworkObject):
            return self.id == __o.id

        return NotImplemented

    def __hash__(self) -> int:
        """
            Calculates hash of the object in a non-randomized manner

            Returns
            -------
            int -> hash of all the items defined on the self.__dict__ object
        """

        return hash(tuple(sorted(self.__dict__.items())))

    def refresh(self, jsonObject: Optional[Dict[str, Any]] = None) -> bool:
        """
            Updates objects fields to a provided value if set, otherwise
            fetches the object from the API and updates the values
            using the fetched object

            Parameters
            ----------
            jsonObject : Optional[Dict[str, Any]]
                A serialized json object to which the values should be updated, if provided

            Returns
            -------
            bool -> True if the update was successful, False otherwise
        """

        # Update from json if it exists
        if jsonObject is not None:
            self._updateFields(jsonObject)
            return True

        # Fetch from server otherwise
        try:
            obj = self.__class__.fetchById(self.id)
        except NetworkRequestError:
            return False

        for key, value in obj.__dict__.items():
            self.__dict__[key] = value

        return True

    def update(self, **kwargs: Any) -> bool:
        """
            Sends a PUT request to Coretex backend

            Parameters
            ----------
            **kwargs : Dict[str, Any]
                parameters which will be sent as request body

            Returns
            -------
            bool -> True if request was successful, False otherwise
        """

        if self.isDeleted:
            return False

        endpoint = f"{self._endpoint()}/{self.id}"
        return not networkManager.put(endpoint, kwargs).hasFailed()

    def delete(self) -> bool:
        """
            Sends a DELETE request to Coretex backend

            Returns
            -------
            bool -> True if request was successful, False otherwise
        """

        if self.isDeleted:
            return False

        endpoint = f"{self._endpoint()}/{self.id}"
        return not networkManager.delete(endpoint).hasFailed()

    @classmethod
    def create(cls, **kwargs: Any) -> Self:
        """
            Creates the entity linked to this class on Coretex backend

            Parameters
            ----------
            **kwargs : Dict[str, Any]
                parameters which will be sent as request body

            Returns
            -------
            Self -> created object if request was successful

            Raises
            ------
            NetworkRequestError -> If the request for creating failed
        """

        response = networkManager.post(cls._endpoint(), kwargs)
        if response.hasFailed():
            raise NetworkRequestError(response, f">> [Coretex] Failed to create \"{cls.__name__}\" with parameters \"{kwargs}\"")

        return cls.decode(response.getJson(dict))

    @classmethod
    def fetchAll(cls, **kwargs: Any) -> List[Self]:
        """
            Fetches all entities from Coretex backend which match
            the given predicate

            Parameters
            ----------
            **kwargs : Optional[Dict[str, Any]]
                query parameters (predicate) which will be appended to URL

            Returns
            -------
            List[Self] -> list of all fetched entities

            Raises
            ------
            NetworkRequestError -> If the request for fetching failed
        """

        if "page_size" not in kwargs:
            kwargs["page_size"] = DEFAULT_PAGE_SIZE

        response = networkManager.get(cls._endpoint(), kwargs)
        if response.hasFailed():
            raise NetworkRequestError(response, f"Failed to fetch \"{cls.__name__}\" with parameters \"{kwargs}\"")

        objects: List[Self] = []

        for obj in response.getJson(list):
            objects.append(cls.decode(obj))

        return objects

    @classmethod
    def fetchOne(cls, **kwargs: Any) -> Self:
        """
            Fetches one entity from Coretex backend which matches
            the given predicate

            Parameters
            ----------
            **kwargs : Optional[Dict[str, Any]]
                query parameters (predicate) which will be appended to URL

            Returns
            -------
            Self -> fetched entity

            Raises
            ------
            NetworkRequestError -> If the request for fetching failed
            ValueError -> If no object was fetched
        """

        kwargs["page_size"] = 1
        result = cls.fetchAll(**kwargs)

        if len(result) == 0:
            raise ValueError(f"Failed to fetch \"{cls.__name__}\" with parameters \"{kwargs}\"")

        return result[0]

    @classmethod
    def fetchById(cls, objectId: int, **kwargs: Any) -> Self:
        """
            Fetches a single entity with the matching id

            Parameters
            ----------
            objectId : int
                id of the object which is fetched
            **kwargs : Optional[Dict[str, Any]]
                query parameters (predicate) which will be appended to URL

            Returns
            -------
            Optional[Self] -> fetched object if request was successful, None otherwise

            Raises
            ------
            NetworkRequestError -> If the request for fetching failed
        """

        if "page_size" not in kwargs:
            kwargs["page_size"] = DEFAULT_PAGE_SIZE

        response = networkManager.get(f"{cls._endpoint()}/{objectId}", kwargs)

        if response.hasFailed():
            raise NetworkRequestError(response, f"Failed to fetch \"{cls.__name__}\" with ID \"{objectId}\"")

        return cls.decode(response.getJson(dict))



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Optional, List, Dict

import os
import xml.etree.ElementTree as ET

from PIL import Image
from skimage import measure
from shapely.geometry import Polygon

import numpy as np

from .shared import getTag, getBoxes
from ....annotation import CoretexSegmentationInstance, BBox
from ....dataset import ImageDataset


ContourPoints = List[List[int]]
SegmentationPolygon = List[ContourPoints]


class ObjectCandidate:

    """
        Represents the groundtruth object for contour matching
    """

    def __init__(self, object: ET.Element):
        self.object = object
        self.matched = False


class ContourCandidate:

    """
        Represents the extracted contour from segmentation mask
    """

    def __init__(self, contours: ContourPoints, iou: Optional[float] = None):
        if iou is None:
            iou = 0.0

        self.contours = contours
        self.iou = iou
        self.matched = False


class InstanceExtractor:

    def __init__(self, dataset: ImageDataset) -> None:
        self.__dataset = dataset

    def createSubmasks(self, maskImage: Image) -> Dict[str, Any]:
        """
            Creates submasks for each segmentation mask

            Parameters
            ----------
            maskImage : Image
                Segmentation mask

            Returns
            -------
            Dict[str, Any] -> Dictionary with submask image and color
        """

        width, height = maskImage.size

        # Initialize a dictionary of sub-masks indexed by RGB colors
        subMasks: Dict[str, Any] = {}
        for x in range(width):
            for y in range(height):
                # Get the RGB values of the pixel
                pixel = maskImage.getpixel((x, y))[:3]

                if pixel == (0, 0, 0):
                    continue

                pixelStr = str(pixel)
                subMask = subMasks.get(pixelStr)
                if subMask is None:
                    subMasks[pixelStr] = Image.new('1', (width + 2, height + 2))

                # Set the pixel value to 1 (default is 0), accounting for padding
                subMasks[pixelStr].putpixel((x + 1, y + 1), 1)

        return subMasks

    def reshapeContour(self, candidate: ContourCandidate) -> ContourPoints:
        listOfPoints: ContourPoints = []

        for segmentContour in candidate.contours:
            for value in range(0, len(segmentContour) - 1, 2):
                point = [segmentContour[value], segmentContour[value + 1]]
                listOfPoints.append(point)

        return listOfPoints

    def calculateIoU(self, boxA: Dict[str, float], boxB: List[float]) -> float:
        """
            Calculates area of overlap for object boxes and contour boxes

            Parameters
            ----------
            boxA : Dict[str, float]
                annotated object boxes
            boxB : Dict[str, float]
                extracted boxes from contour

            Returns
            -------
            Value of boxes area overlap
        """

        xmax = boxA['top_left_x'] + boxA['width']
        ymax = boxA['top_left_y'] + boxA['height']

        xA = max(boxA['top_left_x'], boxB[0])
        yA = max(boxA['top_left_y'], boxB[1])
        xB = min(xmax, boxB[2])
        yB = min(ymax, boxB[3])

        # Compute the area of intersection rectangle
        interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)

        # Compute the area of both the prediction and ground-truth rectangles
        boxAArea = (xmax - boxA['top_left_x'] + 1) * (ymax - boxA['top_left_y'] + 1)
        boxBArea = (boxB[2] - boxB[0] + 1) * (boxB[3] - boxB[1] + 1)

        return interArea / float(boxAArea + boxBArea - interArea)

    def boxesOverlap(self, bbox: Dict[str, Any], listOfPoints: ContourPoints) -> float:
        """
            Extracts boxes from contour and calculates IoU

            Parameters
            ----------
            bbox : Dict[str, Any]
                annotated object boxes
            listOfPoints : ContourPoints
                contour points

            Returns
            -------
            float -> Calculated overlap of boxes area
        """

        xmin: Optional[float] = None
        ymin: Optional[float] = None
        xmax: Optional[float] = None
        ymax: Optional[float] = None

        for point in listOfPoints:
            currentX = point[0]
            currentY = point[1]

            if xmin is None or xmax is None:
                xmin = currentX
                xmax = currentX

            if ymin is None or ymax is None:
                ymin = currentY
                ymax = currentY

            # get xmin, xmax, ymin, ymax
            xmin = min(xmin, currentX)
            ymin = min(ymin, currentY)
            xmax = max(xmax, currentX)
            ymax = max(ymax, currentY)

        if xmax is not None and xmin is not None and ymin is not None and ymax is not None:
            contourBox = [xmin, ymin, xmax, ymax]
            iou = self.calculateIoU(bbox, contourBox)

        return iou

    def matchContour(self, bbox: Dict[str, Any], objectCandidate: ObjectCandidate, contourCandidates: List[ContourCandidate]) -> ContourPoints:
        """
            Matches object and contour with max area of overlap

            Parameters
            ----------
            bbox : Dict[str, Any]
                annotated object boxes
            objectCandidate : ObjectCandidate
                groundtruth object
            contourCandidates : List[ContourCandidate]
                list of ContourCandidate objects

            Returns
            -------
            ContourPoints -> The corresponding contour which matches given object
        """

        maxIoU = -1.0
        contourIndex = 0

        for candidate in contourCandidates:
            if candidate.matched:
                continue

            listOfPoints = self.reshapeContour(candidate)
            candidate.iou = self.boxesOverlap(bbox, listOfPoints)

            if maxIoU < candidate.iou:
                maxIoU = candidate.iou
                contourIndex = contourCandidates.index(candidate)

        objectCandidate.matched = True
        contourCandidates[contourIndex].matched = True

        return contourCandidates[contourIndex].contours

    def extractSubmaskContours(self, subMask: Image) -> ContourPoints:
        """
            Extracts contours from submask image

            Parameters
            ----------
            subMask : Image
                binary image

            Returns
            -------
            ContourPoints -> List of contours
        """

        subMaskArray = np.asarray(subMask)
        contours = measure.find_contours(subMaskArray, 0.5)

        segmentations: ContourPoints = []
        for contour in contours:
            for i in range(len(contour)):
                row, col = contour[i]
                contour[i] = (col - 1, row - 1)

            # Make a polygon and simplify it
            poly = Polygon(contour)
            #poly = poly.simplify(1.0, preserve_topology=False)

            if poly.geom_type == 'MultiPolygon':
                # If MultiPolygon, take the smallest convex Polygon containing all the points in the object
                poly = poly.convex_hull

            # Ignore if still not a Polygon (could be a line or point)
            if poly.geom_type == 'Polygon':
                segmentation = np.array(poly.exterior.coords).ravel().tolist()
                segmentations.append(segmentation)

        return segmentations

    def getSegmentationInstance(
        self,
        objectCandidate: ObjectCandidate,
        contourCandidates: Optional[List[ContourCandidate]]=None
    ) -> Optional[CoretexSegmentationInstance]:

        label = getTag(objectCandidate.object, "name")
        if label is None:
            return None

        clazz = self.__dataset.classByName(label)
        if clazz is None:
            return None

        bndbox = objectCandidate.object.find('bndbox')
        if bndbox is None:
            return None

        boxes = getBoxes(bndbox)
        if boxes is None:
            return None

        polygon: ContourPoints = []
        if contourCandidates is not None:
            polygon = self.matchContour(boxes, objectCandidate, contourCandidates)

        bbox = BBox.decode(boxes)

        if len(polygon) == 0:
            polygon = [bbox.polygon]

        return CoretexSegmentationInstance.create(clazz.classIds[0], bbox, polygon)

    def __extractNonSegmentedInstances(self, objectCandidates: List[ObjectCandidate]) -> List[CoretexSegmentationInstance]:
        coretexInstances: List[CoretexSegmentationInstance] = []

        for objectCandidate in objectCandidates:
            instance = self.getSegmentationInstance(objectCandidate)
            if instance is None:
                continue

            coretexInstances.append(instance)

        return coretexInstances

    def __extractSegmentedInstances(
        self,
        segmentationPath: str,
        filename: str,
        objectCandidates: List[ObjectCandidate]
    ) -> List[CoretexSegmentationInstance]:

        maskImage = Image.open(os.path.join(segmentationPath, filename))
        if maskImage is None:
            return []

        contourCandidates: List[ContourCandidate] = []

        maskImage = maskImage.convert("RGB")
        subMasks = self.createSubmasks(maskImage)

        for color, subMask in subMasks.items():
            # Ignore border contours
            if color == '(224, 224, 192)':
                continue

            annotation = self.extractSubmaskContours(subMask)
            contourCandidates.append(ContourCandidate(annotation))

        coretexInstances: List[CoretexSegmentationInstance] = []

        for objectCandidate in objectCandidates:
            if objectCandidate.matched:
                continue

            instance = self.getSegmentationInstance(objectCandidate, contourCandidates)
            if instance is None:
                continue

            coretexInstances.append(instance)

        return coretexInstances

    def extractInstances(
        self,
        root: ET.Element,
        filename: str,
        segmentationPath: str
    ) -> List[CoretexSegmentationInstance]:
        """
            Extracts polygons from segmentation masks and creates instances

            Parameters
            ----------
            filename : str
                file with annotations
            objects : List[ET.Element]
                annotated objects
        """

        objects = root.findall("object")
        objectCandidates: List[ObjectCandidate] = [ObjectCandidate(obj) for obj in objects]

        segmented = getTag(root, "segmented")
        if segmented is not None:
            isSegmented = bool(int(segmented))

            if isSegmented:
                return self.__extractSegmentedInstances(segmentationPath, filename, objectCandidates)
            else:
                return self.__extractNonSegmentedInstances(objectCandidates)

        # TODO: Raise error or fallback to extracting non-segmented instances?
        raise RuntimeError(">> [Coretex] (segmented) XML tag missing")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Dict
from datetime import datetime

from .project_type import ProjectType
from ...codable import KeyDescriptor
from ...networking import NetworkObject


class BaseObject(NetworkObject):

    """
        Represents the base class for Project/Task objects from Coretex.ai

        Properties
        ----------
        name : str
            name of object
        description : Optional[str]
            description of object
        createdOn : datetime
            date of creation of object
        createdById : str
            id of user that created object
        projectType : ProjectType
            project type of created object
    """

    name: str
    description: Optional[str]
    createdOn: datetime
    createdById: str
    projectType: ProjectType

    @classmethod
    def _endpoint(cls) -> str:
        return "project"

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["projectType"] = KeyDescriptor("project_task", ProjectType)

        return descriptors

    def rename(self, name: str) -> bool:
        """
            Renames the Project/Task

            Parameters
            ----------
            name : str
                new name

            Returns
            -------
            bool -> True if Project/Task was renamed, False if Project/Task was not renamed
        """

        if self.name == name:
            return False

        success = self.update(name = name)

        if success:
            self.name = name

        return success

    def updateDescription(self, description: str) -> bool:
        """
            Updates the Project/Task's description

            Parameters
            ----------
            description : str
                new description

            Returns
                bool -> True if Project/Task's description was updated,
                False if Project/Task's description was not updated
        """

        if self.description == description:
            return False

        success = self.update(description = description)
        if success:
            self.description = description

        return success



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

import click

from ..modules.user import authenticate, saveLoginData
from ..modules.ui import clickPrompt, stdEcho, successEcho
from ...configuration import loadConfig, saveConfig, isUserConfigured


@click.command()
def login() -> None:
    config = loadConfig()
    if isUserConfigured(config):
        if not clickPrompt(
            f"User already logged in with username {config['username']}.\nWould you like to log in with a different user (Y/n)?",
            type = bool,
            default = True,
            show_default = False
        ):
            return

    stdEcho("Please enter your credentials:")
    loginInfo = authenticate()
    config = saveLoginData(loginInfo, config)

    saveConfig(config)

    successEcho(f"User {config['username']} successfully logged in.")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path


RESOURCES_DIR = Path(__file__).resolve().parent



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

import click

from ..modules.user import initializeUserSession
from ..modules.utils import onBeforeCommandExecute
from ..modules.project_utils import getProject
from ... import folder_manager
from ..._task import TaskRunWorker, executeRunLocally, readTaskConfig, runLogger
from ...configuration import loadConfig
from ...entities import TaskRun, TaskRunStatus
from ...resources import PYTHON_ENTRY_POINT_PATH


class RunException(Exception):
    pass


@click.command()
@onBeforeCommandExecute(initializeUserSession)
@click.argument("path", type = click.Path(exists = True, dir_okay = False))
@click.option("--name", type = str, default = None)
@click.option("--description", type = str, default = None)
@click.option("--snapshot", type = bool, default = False)
@click.option("--project", "-p", type = str)
def run(path: str, name: Optional[str], description: Optional[str], snapshot: bool, project: Optional[str]) -> None:
    config = loadConfig()
    parameters = readTaskConfig()

    selectedProject = getProject(project, config)
    if selectedProject is None:
        return

    taskRun: TaskRun = TaskRun.runLocal(
        selectedProject.id,
        snapshot,
        name,
        description,
        [parameter.encode() for parameter in parameters],
        entryPoint = path
    )

    taskRun.updateStatus(TaskRunStatus.preparingToStart)

    with TaskRunWorker(config["refreshToken"], taskRun.id):
        runLogger.attach(taskRun.id)

        command = [
            "python", str(PYTHON_ENTRY_POINT_PATH),
            "--taskRunId", str(taskRun.id),
            "--refreshToken", str(config["refreshToken"])
        ]

        returnCode = executeRunLocally(
            command,
            captureErr = True
        )

    # Flush logs before updating status to a final one
    # as that invalidates the auth token
    runLogger.flushLogs()
    runLogger.reset()

    if returnCode != 0:
        runLogger.fatal(f">> [Coretex] Failed to execute Task. Exit code: {returnCode}")
        taskRun.updateStatus(TaskRunStatus.completedWithError)
    else:
        taskRun.updateStatus(TaskRunStatus.completedWithSuccess)

    folder_manager.clearTempFiles()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TypeVar, Generic, Final
from pathlib import Path

import logging

from .sample import Sample


SampleDataType = TypeVar("SampleDataType")


class LocalSample(Generic[SampleDataType], Sample[SampleDataType]):

    """
        Represents a local sample object\n
        The purpose of this class is to provide a way to work with
        data samples that are stored locally

        Properties
        ----------
        name
            name of sample retrieved from path
        path : Path
            path to local sample
    """

    def __init__(self, path: Path) -> None:
        super().__init__()

        self.name: Final = path.stem
        self._path: Final = path

    @property
    def path(self) -> Path:
        """
            Returns
            -------
            Path -> path for local sample
        """

        return self._path.parent / self._path.stem

    @property
    def zipPath(self) -> Path:
        """
            Returns
            -------
            Path -> zip path for local sample
        """

        return self._path

    def download(self, decrypt: bool = True, ignoreCache: bool = False) -> None:
        logging.getLogger("coretexpylib").warning(">> [Coretex] Local sample cannot be downloaded")

    def load(self) -> SampleDataType:
        return super().load()  # type: ignore
